# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matches BigBrain intesity profiles extracted by Wagstyl et al. to volumes."""

from . import query

from ..features.tabular import bigbrain_intensity_profile, layerwise_bigbrain_intensities
from ..commons import logger
from ..locations import point, pointset
from ..core import region
from ..retrieval import requests, cache
from ..retrieval.datasets import GenericDataset

import numpy as np
from typing import List
from os import path


class WagstylProfileLoader:

    REPO = "https://github.com/kwagstyl/cortical_layers_tutorial/raw/main"
    PROFILES_FILE_LEFT = "data/profiles_left.npy"
    THICKNESSES_FILE_LEFT = "data/thicknesses_left.npy"
    MESH_FILE_LEFT = "data/gray_left_327680.surf.gii"
    _profiles = None
    _vertices = None
    _boundary_depths = None
    DATASET = GenericDataset(
        name="HIBALL workshop on cortical layers",
        contributors=[
            'Konrad Wagstyl',
            'Stéphanie Larocque',
            'Guillem Cucurull',
            'Claude Lepage',
            'Joseph Paul Cohen',
            'Sebastian Bludau',
            'Nicola Palomero-Gallagher',
            'Lindsay B. Lewis',
            'Thomas Funck',
            'Hannah Spitzer',
            'Timo Dickscheid',
            'Paul C. Fletcher',
            'Adriana Romero',
            'Karl Zilles',
            'Katrin Amunts',
            'Yoshua Bengio',
            'Alan C. Evans'
        ],
        url="https://github.com/kwagstyl/cortical_layers_tutorial/",
        description="Cortical profiles of BigBrain staining intensities computed by Konrad Wagstyl, "
        "as described in the publication 'Wagstyl, K., et al (2020). BigBrain 3D atlas of "
        "cortical layers: Cortical and laminar thickness gradients diverge in sensory and "
        "motor cortices. PLoS Biology, 18(4), e3000678. "
        "http://dx.doi.org/10.1371/journal.pbio.3000678."
        "The data is taken from the tutorial at "
        "https://github.com/kwagstyl/cortical_layers_tutorial. Each vertex is "
        "assigned to the regional map when queried."
    )

    def __init__(self):
        if self._profiles is None:
            self.__class__._load()

    @property
    def profile_labels(self):
        return np.arange(0., 1., 1. / self._profiles.shape[1])

    @classmethod
    def _load(cls):
        # read thicknesses, in mm, and normalize by their last column which is the total thickness
        thickness = requests.HttpRequest(f"{cls.REPO}/{cls.THICKNESSES_FILE_LEFT}").data.T
        total_thickness = thickness[:, :-1].sum(1)  # last column is the computed total thickness
        valid = np.where(total_thickness > 0)[0]
        cls._boundary_depths = np.c_[np.zeros_like(valid), (thickness[valid, :] / total_thickness[valid, None]).cumsum(1)]
        cls._boundary_depths[:, -1] = 1  # account for float calculation errors

        # find profiles with valid thickness
        profile_l_url = f"{cls.REPO}/{cls.PROFILES_FILE_LEFT}"
        if not path.exists(cache.CACHE.build_filename(profile_l_url)):
            logger.info(
                "First request to BigBrain profiles. Preprocessing the data "
                "now. This may take a little."
            )
        profiles_l_all = requests.HttpRequest(profile_l_url).data
        cls._profiles = profiles_l_all[valid, :]

        # read mesh vertices
        mesh_left = requests.HttpRequest(f"{cls.REPO}/{cls.MESH_FILE_LEFT}").data
        mesh_vertices = mesh_left.darrays[0].data
        cls._vertices = mesh_vertices[valid, :]

        logger.debug(f"{cls._profiles.shape[0]} BigBrain intensity profiles.")
        assert cls._vertices.shape[0] == cls._profiles.shape[0]

    @staticmethod
    def _choose_space(regionobj: region.Region):
        """
        Helper function to obtain a suitable space to fetch a regional mask.
        BigBrain space has priorty.

        Parameters
        ----------
        regionobj : region.Region

        Returns
        -------
        Space

        Raises
        ------
        RuntimeError
            When there is no supported space that provides image for `regionobj`.
        """
        if regionobj.mapped_in_space('bigbrain'):
            return 'bigbrain'
        supported_spaces = [s for s in regionobj.supported_spaces if s.provides_image]
        if len(supported_spaces) == 0:
            raise RuntimeError(f"Could not find a supported space for {regionobj}")
        return supported_spaces[0]

    def __len__(self):
        return self._vertices.shape[0]

    def match(self, regionobj: region.Region, space: str = None):
        """
        Retrieve BigBrainIntesityProfiles .

        Parameters
        ----------
        regionobj : region.Region
            Region or parcellation
        space : str, optional
            The space in which the region masks will be calculated. By default,
            siibra tries to fetch in BigBrain first and then the other spaces.

        Returns
        -------
        tuple
            tuple of profiles, boundary depths, and vertices
        """
        assert isinstance(regionobj, region.Region)
        logger.debug(f"Matching locations of {len(self)} BigBrain profiles to {regionobj}")

        if space is None:
            space = self._choose_space(regionobj)

        mask = regionobj.fetch_regional_map(space=space, maptype="labelled")
        logger.info(f"Assigning {len(self)} profile locations to {regionobj} in {space}...")
        voxels = (
            pointset.PointSet(self._vertices, space="bigbrain")
            .warp(space)
            .transform(np.linalg.inv(mask.affine), space=None)
        )
        arr = mask.get_fdata()
        XYZ = np.array(voxels.as_list()).astype('int')
        X, Y, Z = np.split(
            XYZ[np.all((XYZ < arr.shape) & (XYZ > 0), axis=1), :],
            3, axis=1
        )
        inside = np.where(arr[X, Y, Z] > 0)[0]

        return (
            self._profiles[inside, :],
            self._boundary_depths[inside, :],
            self._vertices[inside, :]
        )


class BigBrainProfileQuery(query.LiveQuery, args=[], FeatureType=bigbrain_intensity_profile.BigBrainIntensityProfile):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(self, regionobj: region.Region, **kwargs) -> List[bigbrain_intensity_profile.BigBrainIntensityProfile]:
        assert isinstance(regionobj, region.Region)
        loader = WagstylProfileLoader()

        space = WagstylProfileLoader._choose_space(regionobj)
        if not regionobj.is_leaf:
            leaves_defined_on_space = [
                r for r in regionobj.leaves if r.mapped_in_space(space)
            ]
        else:
            leaves_defined_on_space = [regionobj]

        result = []
        for subregion in leaves_defined_on_space:
            matched_profiles, boundary_depths, coords = loader.match(subregion, space)
            for i, profile in enumerate(matched_profiles):
                prof = bigbrain_intensity_profile.BigBrainIntensityProfile(
                    regionname=subregion.name,
                    depths=loader.profile_labels,
                    values=profile,
                    boundaries=boundary_depths[i, :],
                    location=point.Point(coords[i, :], 'bigbrain')  # points are warped into BigBrain
                )
                prof.datasets = [WagstylProfileLoader.DATASET]
                result.append(prof)

        return result


class LayerwiseBigBrainIntensityQuery(query.LiveQuery, args=[], FeatureType=layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities):

    def __init__(self):
        query.LiveQuery.__init__(self)

    def query(self, regionobj: region.Region, **kwargs) -> List[layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities]:
        assert isinstance(regionobj, region.Region)
        loader = WagstylProfileLoader()

        space = WagstylProfileLoader._choose_space(regionobj)
        if not regionobj.is_leaf:
            leaves_defined_on_space = [
                r for r in regionobj.leaves if r.mapped_in_space(space)
            ]
        else:
            leaves_defined_on_space = [regionobj]

        result = []
        for subregion in leaves_defined_on_space:
            matched_profiles, boundary_depths, coords = loader.match(subregion, space)
            if matched_profiles.shape[0] == 0:
                continue

            # compute array of layer labels for all coefficients in profiles_left
            N = matched_profiles.shape[1]
            prange = np.arange(N)
            layer_labels = 7 - np.array([
                [np.array([[(prange < T) * 1] for i, T in enumerate((b * N).astype('int'))]).squeeze().sum(0)]
                for b in boundary_depths
            ]).reshape((-1, 200))

            fp = layerwise_bigbrain_intensities.LayerwiseBigBrainIntensities(
                regionname=subregion.name,
                means=[matched_profiles[layer_labels == layer].mean() for layer in range(1, 7)],
                stds=[matched_profiles[layer_labels == layer].std() for layer in range(1, 7)],
            )
            assert fp.matches(subregion)  # to create an assignment result
            fp.datasets = [WagstylProfileLoader.DATASET]
            result.append(fp)

        return result



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .commons import (
    logger,
    QUIET,
    VERBOSE,
    MapType,
    MapIndex,
    set_log_level,
    __version__
)

from .core import (
    atlas as _atlas,
    parcellation as _parcellation,
    space as _space
)
from .volumes import parcellationmap as _parcellationmap
from .retrieval.requests import (
    EbrainsRequest as _EbrainsRequest,
    CACHE as cache
)
from . import configuration
from .configuration import factory
from . import features, livequeries
from siibra.locations import Point, PointSet

import os as _os
logger.info(f"Version: {__version__}")
logger.warning("This is a development release. Use at your own risk.")
logger.info(
    "Please file bugs and issues at https://github.com/FZJ-INM1-BDA/siibra-python."
)

# forward access to some functions
set_ebrains_token = _EbrainsRequest.set_token
fetch_ebrains_token = _EbrainsRequest.fetch_token
find_regions = _parcellation.Parcellation.find_regions
from_json = factory.Factory.from_json


def __getattr__(attr: str):
    # lazy loading of some classes for package-level functions.
    if attr == 'atlases':
        return _atlas.Atlas.registry()
    elif attr == 'spaces':
        return _space.Space.registry()
    elif attr == 'parcellations':
        return _parcellation.Parcellation.registry()
    elif attr == 'maps':
        return _parcellationmap.Map.registry()
    elif attr == 'use_configuration':
        return configuration.Configuration.use_configuration
    elif attr == 'extend_configuration':
        return configuration.Configuration.extend_configuration
    else:
        raise AttributeError(f"No such attribute: {__name__}.{attr}")


# convenient access to reference space templates
def get_template(space_spec: str, **kwargs):
    return (
        _space.Space
        .get_instance(space_spec)
        .get_template(**kwargs)
    )


# convenient access to parcellation maps
def get_map(parcellation: str, space: str, maptype: MapType = MapType.LABELLED, **kwargs):
    return (
        _parcellation.Parcellation
        .get_instance(parcellation)
        .get_map(space=space, maptype=maptype, **kwargs)
    )


# convenient access to regions of a parcellation
def get_region(parcellation: str, region: str):
    return (
        _parcellation.Parcellation
        .get_instance(parcellation)
        .get_region(region)
    )


def set_feasible_download_size(maxsize_gbyte):
    from .volumes import volume
    volume.gbyte_feasible = maxsize_gbyte
    logger.info(f"Set feasible download size to {maxsize_gbyte} GiB.")


def set_cache_size(maxsize_gbyte: int):
    assert maxsize_gbyte >= 0
    cache.SIZE_GIB = maxsize_gbyte
    logger.info(f"Set cache size to {maxsize_gbyte} GiB.")


if "SIIBRA_CACHE_SIZE_GIB" in _os.environ:
    set_cache_size(float(_os.environ.get("SIIBRA_CACHE_SIZE_GIB")))


def warm_cache():
    """
    Preload preconfigured siibra concepts.

    Siibra relies on preconfigurations that simplify integrating various
    concepts such as parcellations, refernce spaces, and multimodal data
    features. By preloading the instances, siibra commits all preconfigurations
    to the memory at once instead of commiting them when required.
    """
    _ = _atlas.Atlas.registry()
    _ = _space.Space.registry()
    _ = _parcellation.Parcellation.registry()
    _ = _parcellationmap.Map.registry()
    features.warm_cache()
    livequeries.warm_cache()


def __dir__():
    return [
        "atlases",
        "spaces",
        "parcellations",
        "features",
        "use_configuration",
        "extend_configuration",
        "get_region",
        "find_regions",
        "get_map",
        "get_template",
        "MapType",
        "Point",
        "PointSet",
        "QUIET",
        "VERBOSE",
        "fetch_ebrains_token",
        "set_ebrains_token",
        "vocabularies",
        "__version__",
        "cache",
        "warm_cache"
    ]



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles feature queries that rely on live or on-the-fly calculations."""

from ..commons import logger
from ..features.feature import Feature
from ..core.concept import AtlasConcept

from abc import ABC, abstractmethod
from typing import List


class LiveQuery(ABC):

    # set of mandatory query argument names
    _query_args = []

    def __init__(self, **kwargs):
        parstr = ", ".join(f"{k}={v}" for k, v in kwargs.items())
        if parstr:
            parstr = "with parameters " + parstr
        if not all(p in kwargs for p in self._query_args):
            logger.error(
                f"Incomplete specification for {self.__class__.__name__} query "
                f"(Mandatory arguments: {', '.join(self._query_args)})"
            )
        self._kwargs = kwargs

    def __init_subclass__(cls, args: List[str], FeatureType: type):
        cls._query_args = args
        cls.feature_type = FeatureType
        FeatureType._live_queries.append(cls)
        return super().__init_subclass__()

    @abstractmethod
    def query(self, concept: AtlasConcept, **kwargs) -> List[Feature]:
        raise NotImplementedError(f"Dervied class {self.__class__} needs to implement query()")
        pass



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parent class to siibra main concepts."""
from ..commons import (
    create_key,
    clear_name,
    logger,
    InstanceTable,
    Species,
    TypePublication
)

import re
from typing import TypeVar, Type, Union, List, TYPE_CHECKING

T = TypeVar("T", bound="AtlasConcept")


if TYPE_CHECKING:
    from ..retrieval.datasets import EbrainsDataset
    TypeDataset = EbrainsDataset


class AtlasConcept:
    """
    Parent class encapsulating commonalities of the basic siibra concept like atlas, parcellation, space, region.
    These concepts have an id, name, and key, and they are bootstrapped from metadata stored in an online resources.
    Typically, they are linked with one or more datasets that can be retrieved from the same or another online resource,
    providing data files or additional metadata descriptions on request.
    """

    def __init__(
        self,
        identifier: str,
        name: str,
        species: Union[str, Species],
        shortname: str = None,
        description: str = None,
        modality: str = "",
        publications: List[TypePublication] = [],
        datasets: List['TypeDataset'] = [],
        spec=None,
        prerelease: bool = False,
    ):
        """
        Construct a new atlas concept base object.

        Parameters
        ----------
        identifier : str
            Unique identifier of the parcellation
        name : str
            Human-readable name of the parcellation
        species: Species or string
            Specification of the species
        shortname: str
            Shortform of human-readable name (optional)
        description: str
            Textual description of the parcellation
        modality  :  str or None
            Specification of the modality underlying this concept
        datasets : list
            list of datasets corresponding to this concept
        publications: list
            List of publications, each a dictionary with "doi" and/or "citation" fields
        spec: dict, default: None
            The preconfigured specification.
        """
        self._id = identifier
        self.name = name if not prerelease else f"[PRERELEASE] {name}"
        self._species_cached = None if species is None \
            else Species.decode(species)  # overwritable property implementation below
        self.shortname = shortname
        self.modality = modality
        self._description = description
        self._publications = publications
        self.datasets = datasets
        self._spec = spec
        self._prerelease = prerelease

    @property
    def description(self):
        if self._description:
            return self._description
        for ds in self.datasets:
            if ds.description:
                return ds.description
        return ''

    @property
    def LICENSE(self) -> str:
        licenses = []
        for ds in self.datasets:
            if ds.LICENSE is None or ds.LICENSE == "No license information is found.":
                continue
            if isinstance(ds.LICENSE, str):
                licenses.append(ds.LICENSE)
            if isinstance(ds.LICENSE, list):
                licenses.extend(ds.LICENSE)
        return '\n'.join(licenses)

    @property
    def doi_or_url(self) -> str:
        return '\n'.join([
            url.get("url")
            for ds in self.datasets
            for url in ds.urls
        ])

    @property
    def authors(self):
        return [
            contributer['name']
            for ds in self.datasets
            for contributer in ds.contributors
        ]

    @property
    def publications(self) -> List[TypePublication]:
        return [
            *self._publications,
            *[
                {'citation': f"Dataset name: {ds.name}", 'url': url.get("url")}
                for ds in self.datasets
                for url in ds.urls
            ]
        ]

    @property
    def species(self) -> Species:
        # Allow derived classes to implement a lazy loader (e.g. in Map)
        if self._species_cached is None:
            raise RuntimeError(f"No species defined for {self}.")
        return self._species_cached

    @classmethod
    def registry(cls: Type[T]) -> InstanceTable[T]:
        if cls._configuration_folder is None:
            return None
        if cls._registry_cached is None:
            from ..configuration import Configuration
            conf = Configuration()
            # visit the configuration to provide a cleanup function
            # in case the user changes the configuration during runtime.
            Configuration.register_cleanup(cls.clear_registry)
            assert cls._configuration_folder in conf.folders
            objects = conf.build_objects(cls._configuration_folder)
            logger.debug(f"Built {len(objects)} preconfigured {cls.__name__} objects.")
            assert len(objects) > 0
            assert all([hasattr(o, 'key') for o in objects])

            # TODO Map.registry() returns InstanceTable that contains two different types, SparseMap and Map
            # Since we take the objects[0].__class__.match, if the first element happen to be SparseMap, this could result.
            # Code to reproduce:
            """
            import siibra
            r = siibra.volumes.Map.registry()
            """
            if len({o.__class__ for o in objects}) > 1:
                logger.warning(f"{cls.__name__} registry contains multiple classes: {', '.join(list({o.__class__.__name__ for o in objects}))}")
            assert hasattr(objects[0].__class__, "match") and callable(objects[0].__class__.match)
            cls._registry_cached = InstanceTable(
                elements={o.key: o for o in objects},
                matchfunc=objects[0].__class__.match
            )
        return cls._registry_cached

    @classmethod
    def clear_registry(cls):
        cls._registry_cached = None

    @classmethod
    def get_instance(cls, spec: str):
        """
        Parameters
        ----------
            spec: str
                Specification of the class the instance is requested.
        Returns
        -------
            an instance of this class matching the given specification from its
            registry if possible, otherwise None.
        Raises
        ------
            IndexError
                If spec cannot match any instance
        """
        if cls.registry() is not None:
            return cls.registry().get(spec)

    @property
    def id(self):
        # allows derived classes to assign the id dynamically
        return self._id

    @property
    def key(self):
        return create_key(self.name)

    def __init_subclass__(cls, configuration_folder: str = None):
        """
        This method is called whenever AtlasConcept gets subclassed
        (see https://docs.python.org/3/reference/datamodel.html)
        """
        cls._registry_cached = None
        cls._configuration_folder = configuration_folder
        return super().__init_subclass__()

    def __str__(self):
        return f"{self.__class__.__name__}: {self.name}"

    def matches(self, spec):
        """
        Parameters
        ----------
            spec: str
                Specification checked within the concept name, key or id
        Returns
        -------
            bool
                Whether the given specification matches the name, key or id of the concept.
        """
        if isinstance(spec, self.__class__) and (spec == self):
            return True
        elif isinstance(spec, str):
            if spec == self.key:
                return True
            elif spec == self.id:
                return True
            else:
                # match the name
                words = [w for w in re.split("[ -]", spec)]
                squeezedname = clear_name(self.name.lower()).replace(" ", "")
                return any(
                    [
                        all(w.lower() in squeezedname for w in words),
                        spec.replace(" ", "") in squeezedname,
                    ]
                )
        return False

    @classmethod
    def match(cls, obj, spec):
        """Match a given object specification. """
        assert isinstance(obj, cls)
        return obj.matches(spec)

    def __gt__(self, other: 'AtlasConcept'):
        """
        Compare this atlas concept with other atlas concepts of the same kind
        with it's name.
        """
        if self.__class__ is not other.__class__:
            raise ValueError("Cannot compare different atlas concept types.")
        return self.name > other.name

    def __lt__(self, other: 'AtlasConcept'):
        """
        Compare this atlas concept with other atlas concepts of the same kind
        with it's name.
        """
        if self.__class__ is not other.__class__:
            raise ValueError("Cannot compare different atlas concept types.")
        return self.name < other.name



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zipfile import ZipFile
from ..feature import Feature
from ..tabular.tabular import Tabular

from .. import anchor as _anchor

from ...commons import logger, QUIET, siibra_tqdm
from ...core import region as _region
from ...locations import pointset
from ...retrieval.repositories import RepositoryConnector

import pandas as pd
import numpy as np
from typing import Callable, Dict, Union, List

try:
    from typing import Literal
except ImportError:  # support python 3.7
    from typing_extensions import Literal


class RegionalConnectivity(Feature):
    """
    Parcellation-averaged connectivity, providing one or more matrices of a
    given modality for a given parcellation.
    """

    def __init__(
        self,
        cohort: str,
        modality: str,
        regions: list,
        connector: RepositoryConnector,
        decode_func: Callable,
        files: Dict[str, str],
        anchor: _anchor.AnatomicalAnchor,
        description: str = "",
        datasets: list = [],
        prerelease: bool = False,
        id: str = None,
    ):
        """
        Construct a parcellation-averaged connectivity matrix.

        Parameters
        ----------
        cohort: str
            Name of the cohort used for computing the connectivity.
        modality: str
            Connectivity modality, typically set by derived classes.
        regions: list[str]
            Names of the regions from the parcellation
        connector: Connector
            Repository connector for loading the actual data array(s).
        decode_func: function
            Function to convert the bytestream of a loaded file into an array
        files: dict
            A dictionary linking names of matrices (typically subject ids)
            to the relative filenames of the data array(s) in the repository connector.
        anchor: AnatomicalAnchor
            anatomical localization of the matrix, expected to encode the parcellation
            in the region attribute.
        description: str, optional
            textual description of this connectivity matrix.
        datasets : list[Dataset]
            list of datasets corresponding to this feature.
        """
        Feature.__init__(
            self,
            modality=modality,
            description=description or '\n'.join({ds.description for ds in datasets}),
            anchor=anchor,
            datasets=datasets,
            prerelease=prerelease,
            id=id,
        )
        self.cohort = cohort.upper()
        self._connector = connector
        self._files = files
        self._decode_func = decode_func
        self.regions = regions
        self._matrices = {}

    @property
    def subjects(self):
        """
        Returns the subject identifiers for which matrices are available.
        """
        return list(self._files.keys())

    @property
    def name(self):
        supername = super().name
        return f"{supername} with cohort {self.cohort}"

    def get_matrix(self, subject: str = None):
        """
        Returns a matrix as a pandas dataframe.

        Parameters
        ----------
        subject: str, default: None
            Name of the subject (see ConnectivityMatrix.subjects for available names).
            If None, the mean is taken in case of multiple available matrices.
        Returns
        -------
        pd.DataFrame
            A square matrix with region names as the column and row names.
        """
        assert len(self) > 0
        if (subject is None) and (len(self) > 1):
            # multiple matrices available, but no subject given - return mean matrix
            logger.info(
                f"No subject name supplied, returning mean connectivity across {len(self)} subjects. "
                "You might alternatively specify an individual subject."
            )
            if "mean" not in self._matrices:
                all_arrays = [
                    self._connector.get(fname, decode_func=self._decode_func)
                    for fname in siibra_tqdm(
                        self._files.values(),
                        total=len(self),
                        desc=f"Averaging {len(self)} connectivity matrices"
                    )
                ]
                self._matrices['mean'] = self._arraylike_to_dataframe(np.stack(all_arrays).mean(0))
            return self._matrices['mean'].copy()
        if subject is None:
            subject = next(iter(self._files.keys()))
        if subject not in self._files:
            raise ValueError(f"Subject name '{subject}' not known, use one of: {', '.join(self._files)}")
        if subject not in self._matrices:
            self._matrices[subject] = self._load_matrix(subject)
        return self._matrices[subject].copy()

    def plot_matrix(
        self, subject: str = None, regions: List[str] = None,
        logscale: bool = False, *args, backend="nilearn", **kwargs
    ):
        """
        Plots the heatmap of the connectivity matrix using nilearn.plotting.

        Parameters
        ----------
        subject: str
            Name of the subject (see ConnectivityMatrix.subjects for available names).
            If "mean" or None is given, the mean is taken in case of multiple
            available matrices.
        regions: list[str]
            Display the matrix only for selected regions. By default, shows all the regions.
            It can only be a subset of regions of the feature.
        logscale: bool
            Display the data in log10 scale
        backend: str
            "nilearn" or "plotly"
        **kwargs:
            Can take all the arguments `nilearn.plotting.plot_matrix` can take. See the doc at
            https://nilearn.github.io/stable/modules/generated/nilearn.plotting.plot_matrix.html
        """
        if regions is None:
            regions = self.regions
        indices = [self.regions.index(r) for r in regions]
        matrix = self.get_matrix(subject=subject).iloc[indices, indices].to_numpy()  # nilearn.plotting.plot_matrix works better with a numpy array

        if logscale:
            matrix = np.log10(matrix)

        # default kwargs
        subject_title = subject or ""
        kwargs["title"] = kwargs.get(
            "title",
            f"{subject_title} - {self.modality} in {', '.join({_.name for _ in self.anchor.regions})}"
        )

        if kwargs.get("reorder") or (backend == "nilearn"):
            kwargs["figure"] = kwargs.get("figure", (15, 15))
            from nilearn import plotting
            plotting.plot_matrix(
                matrix,
                labels=regions,
                **kwargs
            )
        elif backend == "plotly":
            from plotly.express import imshow
            return imshow(matrix, *args, x=regions, y=regions, **kwargs)
        else:
            raise NotImplementedError(
                f"Plotting connectivity matrices with {backend} is not supported."
            )

    def __iter__(self):
        return ((sid, self.get_matrix(sid)) for sid in self._files)

    def _export(self, fh: ZipFile):
        super()._export(fh)
        for sub in self.subjects:
            df = self.get_matrix(sub)
            fh.writestr(f"sub/{sub}/matrix.csv", df.to_csv())

    def get_profile(
        self,
        region: Union[str, _region.Region],
        subject: str = None,
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column'
    ):
        """
        Extract a regional profile from the matrix, to obtain a tabular data
        feature with the connectivity as the single column. Rows are be sorted
        by descending connection strength.

        Parameters
        ----------
        region: str, Region
        subject: str, default: None
        min_connectivity: float, default: 0
            Regions with connectivity less than this value are discarded.
        max_rows: int, default: None
            Max number of regions with highest connectivity.
        direction: str, default: 'column'
            Choose the direction of profile extraction particularly for
            non-symmetric matrices. ('column' or 'row')
        """
        matrix = self.get_matrix(subject)
        if direction.lower() not in ['column', 'row']:
            raise ValueError("Direction can only be 'column' or 'row'")
        if direction.lower() == 'row':
            matrix = matrix.transpose()

        def matches(r1, r2):
            if isinstance(r1, tuple):
                return any(r.matches(r2) for r in r1)
            else:
                assert isinstance(r1, _region.Region)
                return r1.matches(r2)

        regions = [r for r in matrix.index if matches(r, region)]
        if len(regions) == 0:
            raise ValueError(f"Invalid region specificiation: {region}")
        elif len(regions) > 1:
            raise ValueError(f"Region specification {region} matched more than one profile: {regions}")
        else:
            name = \
                f"Averaged {self.modality}" if subject is None \
                else f"{self.modality}"
            series = matrix[regions[0]]
            last_index = len(series) - 1 if max_rows is None \
                else min(max_rows, len(series) - 1)
            return Tabular(
                description=self.description,
                modality=f"{self.modality} {self.cohort}",
                anchor=_anchor.AnatomicalAnchor(
                    species=list(self.anchor.species)[0],
                    region=regions[0]
                ),
                data=(
                    series[:last_index]
                    .to_frame(name=name)
                    .query(f'`{name}` > {min_connectivity}')
                    .sort_values(by=name, ascending=False)
                    .rename_axis('Target regions')
                ),
                datasets=self.datasets
            )

    def plot_profile(
        self,
        region: Union[str, _region.Region],
        subject: str = None,
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column',
        logscale: bool = False,
        *args,
        backend="matplotlib",
        **kwargs
    ):
        profile = self.get_profile(region, subject, min_connectivity, max_rows, direction)
        kwargs["kind"] = kwargs.get("kind", "barh")
        if backend == "matplotlib":
            kwargs["logx"] = kwargs.get("logx", logscale)
        elif backend == "plotly":
            kwargs.update({
                "color": kwargs.get("color", profile.data.columns[0]),
                "x": kwargs.get("x", profile.data.columns[0]),
                "y": kwargs.get("y", [r.name for r in profile.data.index]),
                "log_x": logscale,
                "labels": {"y": " ", "x": ""},
                "color_continuous_scale": "jet",
                "width": 600, "height": 3800
            })
            fig = profile.data.plot(*args, backend=backend, **kwargs)
            fig.update_layout({
                "font": dict(size=9),
                "yaxis": {"autorange": "reversed"},
                "coloraxis": {"colorbar": {
                    "orientation": "h", "title": "", "xpad": 0, "ypad": 10
                }},
                "margin": dict(l=0, r=0, b=0, t=0, pad=0)
            })
            return fig
        return profile.plot(*args, backend=backend, **kwargs)

    def __len__(self):
        return len(self._files)

    def __str__(self):
        return "{} connectivity for {} from {} cohort ({} matrices)".format(
            self.paradigm if hasattr(self, "paradigm") else self.modality,
            "_".join(p.name for p in self.anchor.parcellations),
            self.cohort,
            len(self._files),
        )

    def compute_centroids(self, space):
        """
        Computes the list of centroid coordinates corresponding to
        matrix rows, in the given reference space.

        Parameters
        ----------
        space: Space, str
        Returns
        -------
        list[tuple(float, float, float)]
        """
        result = []
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parcmap = next(iter(parcellations)).get_map(space)
        all_centroids = parcmap.compute_centroids()
        for regionname in self.regions:
            region = parcmap.parcellation.get_region(regionname, allow_tuple=True)
            if isinstance(region, tuple):  # deal with sets of matched regions
                found = [c for r in region for c in r if c.name in all_centroids]
            else:
                found = [r for r in region if r.name in all_centroids]
            assert len(found) > 0
            result.append(
                tuple(pointset.PointSet(
                    [all_centroids[r.name] for r in found], space=space
                ).centroid)
            )
        return result

    def _arraylike_to_dataframe(self, array: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        """
        Convert a numpy array with the connectivity matrix to
        a DataFrame with regions as column and row headers.
        """
        if not isinstance(array, np.ndarray):
            array = array.to_numpy()
        if not (array == array.T).all():
            logger.warning("The connectivity matrix is not symmetric.")
        df = pd.DataFrame(array)
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parc = next(iter(parcellations))
        with QUIET:
            indexmap = {
                i: parc.get_region(regionname, allow_tuple=True)
                for i, regionname in enumerate(self.regions)
            }
        nrows = array.shape[0]
        try:
            assert len(indexmap) == nrows
            remapper = {
                label - min(indexmap.keys()): region
                for label, region in indexmap.items()
            }
            df = df.rename(index=remapper).rename(columns=remapper)
        except Exception:
            raise RuntimeError("Could not decode connectivity matrix regions.")
        return df

    def _load_matrix(self, subject: str):
        """
        Extract connectivity matrix.
        """
        assert subject in self.subjects
        array = self._connector.get(self._files[subject], decode_func=self._decode_func)
        nrows = array.shape[0]
        if array.shape[1] != nrows:
            raise RuntimeError(
                f"Non-quadratic connectivity matrix {nrows}x{array.shape[1]} "
                f"from {self._files[subject]} in {str(self._connector)}"
            )
        return self._arraylike_to_dataframe(array)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from textwrap import wrap
from zipfile import ZipFile
from io import BytesIO
from os import path

from ..feature import Feature
from ..tabular import tabular

from .. import anchor as _anchor

from ...commons import logger, QUIET, siibra_tqdm
from ...core import region as _region
from ...locations import pointset
from ...retrieval.repositories import RepositoryConnector, ZipfileConnector
from ...retrieval.cache import CACHE

import pandas as pd
import numpy as np
from typing import Callable, Dict, Union, List

try:
    from typing import Literal
except ImportError:  # support python 3.7
    from typing_extensions import Literal

def name_to_code(name):
    map_ = {
        "areas 1 and 2 of cortex": "A1-2",
        "area 10 of cortex": "A10",
        "area 11 of cortex": "A11",
        "area 13 of cortex, lateral part": "A13L",
        "area 13 of cortex, medial part": "A13M",
        "area 13a of cortex": "A13a",
        "area 13b of cortex": "A13b",
        "area 14 of cortex, caudal part": "A14C",
        "area 14 of cortex, rostral part": "A14R",
        "area 19 of cortex, dorsointermediate part": "A19DI",
        "area 19 of cortex, medial part": "A19M",
        "area 23 of cortex, ventral part": "A23V",
        "area 23a of cortex": "A23a",
        "area 23b of cortex": "A23b",
        "area 23c of cortex": "A23c",
        "area 24a of cortex": "A24a",
        "area 24b of cortex": "A24b",
        "area 24c of cortex": "A24c",
        "area 24d of cortex": "A24d",
        "area 25 of cortex": "A25",
        "area 29a-c of cortex": "A29a-c",
        "area 29d of cortex": "A29d",
        "area 30 of cortex": "A30",
        "area 31 of cortex": "A31",
        "area 32 of cortex": "A32",
        "area 32 of cortex, ventral part": "A32V",
        "area 35 of cortex": "A35",
        "area 36 of cortex": "A36",
        "area 3a of cortex (somatosensory)": "A3a",
        "area 3b of cortex (somatosensory)": "A3b",
        "area 45 of cortex": "A45",
        "area 46 of cortex, dorsal part": "A46D",
        "area 46 of cortex, ventral part": "A46V",
        "area 47 (old 12) of cortex, lateral part": "A47L",
        "area 47 (old 12) of cortex, medial part": "A47M",
        "area 47 (old 12) of cortex, orbital part": "A47O",
        "area 4 of cortex, parts a and b (primary motor)": "A4ab",
        "area 4 of cortex, part c (primary motor)": "A4c",
        "area 6 of cortex, dorsocaudal part": "A6DC",
        "area 6 of cortex, dorsorostral part": "A6DR",
        "area 6 of cortex, medial (supplementary motor) part": "A6M",
        "area 6 of cortex, ventral, part a": "A6Va",
        "area 6 of cortex, ventral, part b": "A6Vb",
        "area 8 of cortex, caudal part": "A8C",
        "area 8a of cortex, dorsal part": "A8aD",
        "area 8a of cortex, ventral part": "A8aV",
        "area 8b of cortex": "A8b",
        "area 9 of cortex": "A9",
        "agranular insular cortex": "AI",
        "anterior intraparietal area of cortex": "AIP",
        "amygdalopiriform transition area": "APir",
        "auditory cortex, primary area": "AuA1",
        "auditory cortex, anterolateral area": "AuAL",
        "auditory cortex, caudolateral area": "AuCL",
        "auditory cortex, caudomedial area": "AuCM",
        "auditory cortex, caudal parabelt area": "AuCPB",
        "auditory cortex, middle lateral area": "AuML",
        "auditory cortex, rostral area": "AuR",
        "auditory cortex, rostromedial area": "AuRM",
        "auditory cortex, rostral parabelt": "AuRPB",
        "auditory cortex, rostrotemporal": "AuRT",
        "auditory cortex, rostrotemporal lateral area": "AuRTL",
        "auditory cortex, rostrotemporal medial area": "AuRTM",
        "dysgranular insular cortex": "DI",
        "entorhinal cortex": "Ent",
        "fundus of superior temporal sulcus area of cortex": "FST",
        "granular insular cortex": "GI",
        "gustatory cortex": "Gu",
        "insular proisocortex": "IPro",
        "lateral intraparietal area of cortex": "LIP",
        "medial intraparietal area of cortex": "MIP",
        "medial superior temporal area of cortex": "MST",
        "orbital periallocortex": "OPAl",
        "orbital proisocortex": "OPro",
        "occipito-parietal transitional area of cortex": "OPt",
        "parietal area PE": "PE",
        "parietal area PE, caudal part": "PEC",
        "parietal area PF (cortex)": "PF",
        "parietal area PFG (cortex)": "PFG",
        "parietal area PG": "PG",
        "parietal area PG, medial part (cortex)": "PGM",
        "parietal areas PGa and IPa (fundus of superior temporal ventral area)": "PGa-IPa",
        "parainsular cortex, lateral part": "PaIL",
        "parainsular cortex, medial part": "PaIM",
        "piriform cortex": "Pir",
        "proisocortical motor region (precentral opercular cortex)": "ProM",
        "prostriate area": "ProSt",
        "retroinsular area (cortex)": "ReI",
        "secondary somatosensory cortex, external part": "S2E",
        "secondary somatosensory cortex, internal part": "S2I",
        "secondary somatosensory cortex, parietal rostral area": "S2PR",
        "secondary somatosensory cortex, parietal ventral area": "S2PV",
        "superior temporal rostral area (cortex)": "STR",
        "temporal area TE1 (inferior temporal cortex)": "TE1",
        "temporal area TE2 (inferior temporal cortex)": "TE2",
        "temporal area TE3 (inferior temporal cortex)": "TE3",
        "temporal area TE, occipital part": "TEO",
        "temporal area TF": "TF",
        "temporal area TF, occipital part": "TFO",
        "temporal area TH": "TH",
        "temporal area TL": "TL",
        "temporal area TL, occipital part": "TLO",
        "temporo-parieto-occipital association area (superior temporal polysensory cortex)": "TPO",
        "temporopolar proisocortex": "TPPro",
        "temporal proisocortex": "TPro",
        "temporoparietal transitional area": "TPt",
        "primary visual cortex": "V1",
        "visual area 2": "V2",
        "visual area 3 (ventrolateral posterior area)": "V3",
        "visual area 3A (dorsoanterior area)": "V3A",
        "visual area 4 (ventrolatereral anterior area)": "V4",
        "visual area 4, transitional part (middle temporal crescent)": "V4T",
        "visual area 5 (middle temporal area)": "V5",
        "visual area 6 (dorsomedial area)": "V6",
        "visual area 6A (posterior parietal medial area)": "V6A",
        "ventral intraparietal area of cortex": "VIP"
    }
    return map_[name]


class DFWithMeta(pd.DataFrame):
    _metadata = ['meta']
    @property
    def _constructor(self):
        return DFWithMeta

class InterarealConnectivityMatrix(
    tabular.Tabular,
    configuration_folder="features/tabular/connectivitystrength",
    category="cellular",
):
    """
    Parcellation-averaged connectivity, providing one or more matrices of a
    given modality for a given parcellation.
    """

    @classmethod
    def decode_meta(cls, spec):
        decoder_spec = spec.get("decoder", {})
        if decoder_spec["@type"].endswith('csv'):
            kwargs = {k: v for k, v in decoder_spec.items() if k != "@type"}
            return lambda b: pd.read_csv(BytesIO(b), **kwargs)
        else:
            return None


    class ConnectivityConnector(ZipfileConnector):

        class ZipFileLoader:
            """
            Loads a file from the zip archive, but mimics the behaviour
            of cached http requests used in other connectors.
            """
            def __init__(self, zipfile, filename, decode_func, meta=None):
                self.zipfile = zipfile
                self.filename = filename
                self.func = decode_func
                self.cachefile = CACHE.build_filename(zipfile + filename)
                self.meta = meta

            @property
            def cached(self):
                return path.isfile(self.cachefile)

            @property
            def data(self):
                container = ZipFile(self.zipfile)
                df = self.func(container.open(self.filename).read())
                if self.meta is not None:
                    df = DFWithMeta(df)
                    df.meta = self.meta
                return df


        def get_loader(self, filename, folder="", decode_func=None):
            """Get a lazy loader for a file, for loading data
            only once loader.data is accessed."""
            meta = self.ZipFileLoader(self.zipfile, 'meta.json', lambda b: json.loads(b)).data
            loader = self.ZipFileLoader(self.zipfile, filename, decode_func, meta)
            return loader

    def __init__(
        self,
        cohort: str,
        modality: str,
        regions: list,
        connector: RepositoryConnector,
        decode_func: Callable,
        files: Dict[str, str],
        anchor: _anchor.AnatomicalAnchor,
        description: str = "",
        datasets: list = [],
        prerelease: bool = False,
        id: str = None,
    ):
        """
        Construct a parcellation-averaged connectivity matrix.

        Parameters
        ----------
        cohort: str
            Name of the cohort used for computing the connectivity.
        modality: str
            Connectivity modality, typically set by derived classes.
        regions: list[str]
            Names of the regions from the parcellation
        connector: Connector
            Repository connector for loading the actual data array(s).
        decode_func: function
            Function to convert the bytestream of a loaded file into an array
        files: dict
            A dictionary linking names of matrices (typically subject ids)
            to the relative filenames of the data array(s) in the repository connector.
        anchor: AnatomicalAnchor
            anatomical localization of the matrix, expected to encode the parcellation
            in the region attribute.
        description: str, optional
            textual description of this connectivity matrix.
        datasets : list[Dataset]
            list of datasets corresponding to this feature.
        """
        tabular.Tabular.__init__(
            self,
            modality=modality,
            description=description or '\n'.join({ds.description for ds in datasets}),
            anchor=anchor,
            datasets=datasets,
            data=None,  # lazy loading below
            prerelease=prerelease,
            id=id,
        )
        self.cohort = cohort.upper()
        self._connector = connector
        self._files = files
        if decode_func is not None:
            self._decode_func = decode_func
        else:
            self._decode_func = self.decode_meta
        self.regions = regions
        self._matrices = {}
    @property
    def subjects(self):
        """
        Returns the subject identifiers for which matrices are available.
        """
        return list(self._files.keys())

    @property
    def name(self):
        supername = super().name
        return f"{supername} with cohort {self.cohort}"

    def get_matrix(self, subject: str = None):
        """
        Returns a matrix as a pandas dataframe.

        Parameters
        ----------
        subject: str, default: None
            Name of the subject (see ConnectivityMatrix.subjects for available names).
            If None, the mean is taken in case of multiple available matrices.
        Returns
        -------
        pd.DataFrame
            A square matrix with region names as the column and row names.
        """
        assert len(self) > 0
        if (subject is None) and (len(self) > 1):
            # multiple matrices available, but no subject given - return mean matrix
            logger.info(
                f"No subject name supplied, returning mean connectivity across {len(self)} subjects. "
                "You might alternatively specify an individual subject."
            )
            if "mean" not in self._matrices:
                all_arrays = [
                    self._connector.get(fname, decode_func=self._decode_func)
                    for fname in siibra_tqdm(
                        self._files.values(),
                        total=len(self),
                        desc=f"Averaging {len(self)} connectivity matrices"
                    )
                ]
                self._matrices['mean'] = self._arraylike_to_dataframe(np.stack(all_arrays).mean(0))
            return self._matrices['mean'].copy()

        if subject is None:
            subject = next(iter(self._files.keys()))
        if subject not in self._files:
            raise ValueError(f"Subject name '{subject}' not known, use one of: {', '.join(self._files)}")
        if subject not in self._matrices:
            self._matrices[subject] = self._load_matrix(subject)
        return self._matrices[subject].copy()

    def plot_matrix(
        self, subject: str = None, regions: List[str] = None,
        logscale: bool = False, *args, backend="nilearn", **kwargs
    ):
        """
        Plots the heatmap of the connectivity matrix using nilearn.plotting.

        Parameters
        ----------
        subject: str
            Name of the subject (see ConnectivityMatrix.subjects for available names).
            If "mean" or None is given, the mean is taken in case of multiple
            available matrices.
        regions: list[str]
            Display the matrix only for selected regions. By default, shows all the regions.
            It can only be a subset of regions of the feature.
        logscale: bool
            Display the data in log10 scale
        backend: str
            "nilearn" or "plotly"
        **kwargs:
            Can take all the arguments `nilearn.plotting.plot_matrix` can take. See the doc at
            https://nilearn.github.io/stable/modules/generated/nilearn.plotting.plot_matrix.html
        """
        if regions is None:
            regions = self.regions
        indices = [self.regions.index(r) for r in regions]
        matrix = self.get_matrix(subject=subject).iloc[indices, indices].to_numpy()  # nilearn.plotting.plot_matrix works better with a numpy array

        if logscale:
            matrix = np.log10(matrix)

        # default kwargs
        subject_title = subject or ""
        kwargs["title"] = kwargs.get(
            "title",
            f"{subject_title} - {self.modality} in {', '.join({_.name for _ in self.anchor.regions})}"
        )

        if kwargs.get("reorder") or (backend == "nilearn"):
            kwargs["figure"] = kwargs.get("figure", (15, 15))
            from nilearn import plotting
            plotting.plot_matrix(
                matrix,
                labels=regions,
                **kwargs
            )
        elif backend == "plotly":
            from plotly.express import imshow
            return imshow(matrix, *args, x=regions, y=regions, **kwargs)
        else:
            raise NotImplementedError(
                f"Plotting connectivity matrices with {backend} is not supported."
            )

    def __iter__(self):
        return ((sid, self.get_matrix(sid)) for sid in self._files)

    def _export(self, fh: ZipFile):
        super()._export(fh)
        for sub in self.subjects:
            df = self.get_matrix(sub)
            fh.writestr(f"sub/{sub}/matrix.csv", df.to_csv())

    def get_profile(
        self,
        region: Union[str, _region.Region],
        subject: str = None,
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column'
    ):
        """
        Extract a regional profile from the matrix, to obtain a tabular data
        feature with the connectivity as the single column. Rows are be sorted
        by descending connection strength.

        Parameters
        ----------
        region: str, Region
        subject: str, default: None
        min_connectivity: float, default: 0
            Regions with connectivity less than this value are discarded.
        max_rows: int, default: None
            Max number of regions with highest connectivity.
        direction: str, default: 'column'
            Choose the direction of profile extraction particularly for
            non-symmetric matrices. ('column' or 'row')
        """
        matrix = self.get_matrix(subject)
        if direction.lower() not in ['column', 'row']:
            raise ValueError("Direction can only be 'column' or 'row'")
        if direction.lower() == 'row':
            matrix = matrix.transpose()

        def matches(r1, r2):
            if isinstance(r1, tuple):
                return any(r.matches(r2) for r in r1)
            else:
                assert isinstance(r1, _region.Region)
                return r1.matches(r2)

        regions = [r for r in matrix.index if matches(r, region)]
        if len(regions) == 0:
            raise ValueError(f"Invalid region specificiation: {region}")
        elif len(regions) > 1:
            raise ValueError(f"Region specification {region} matched more than one profile: {regions}")
        else:
            name = \
                f"Averaged {self.modality}" if subject is None \
                else f"{self.modality}"
            series = matrix[regions[0]]
            last_index = len(series) - 1 if max_rows is None \
                else min(max_rows, len(series) - 1)
            return tabular.Tabular(
                description=self.description,
                modality=f"{self.modality} {self.cohort}",
                anchor=_anchor.AnatomicalAnchor(
                    species=list(self.anchor.species)[0],
                    region=regions[0]
                ),
                data=(
                    series[:last_index]
                    .to_frame(name=name)
                    .query(f'`{name}` > {min_connectivity}')
                    .sort_values(by=name, ascending=False)
                    .rename_axis('Target regions')
                ),
                datasets=self.datasets
            )

    def plot_profile(
        self,
        region: Union[str, _region.Region],
        subject: str = None,
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column',
        logscale: bool = False,
        *args,
        backend="matplotlib",
        **kwargs
    ):
        profile = self.get_profile(region, subject, min_connectivity, max_rows, direction)
        kwargs["kind"] = kwargs.get("kind", "barh")
        if backend == "matplotlib":
            kwargs["logx"] = kwargs.get("logx", logscale)
        elif backend == "plotly":
            kwargs.update({
                "color": kwargs.get("color", profile.data.columns[0]),
                "x": kwargs.get("x", profile.data.columns[0]),
                "y": kwargs.get("y", [r.name for r in profile.data.index]),
                "log_x": logscale,
                "labels": {"y": " ", "x": ""},
                "color_continuous_scale": "jet",
                "width": 600, "height": 3800
            })
            fig = profile.data.plot(*args, backend=backend, **kwargs)
            fig.update_layout({
                "font": dict(size=9),
                "yaxis": {"autorange": "reversed"},
                "coloraxis": {"colorbar": {
                    "orientation": "h", "title": "", "xpad": 0, "ypad": 10
                }},
                "margin": dict(l=0, r=0, b=0, t=0, pad=0)
            })
            return fig
        return profile.plot(*args, backend=backend, **kwargs)

    def __len__(self):
        return len(self._files)

    @property
    def data(self):
        m = self.get_matrix(subject=None)
        return m

    def __str__(self):
        return "{} connectivity for {} from {} cohort ({} matrices)".format(
            self.paradigm if hasattr(self, "paradigm") else self.modality,
            "_".join(p.name for p in self.anchor.parcellations),
            self.cohort,
            len(self._files),
        )

    def compute_centroids(self, space):
        """
        Computes the list of centroid coordinates corresponding to
        matrix rows, in the given reference space.

        Parameters
        ----------
        space: Space, str
        Returns
        -------
        list[tuple(float, float, float)]
        """
        result = []
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parcmap = next(iter(parcellations)).get_map(space)
        all_centroids = parcmap.compute_centroids()
        for regionname in self.regions:
            region = parcmap.parcellation.get_region(regionname, allow_tuple=True)
            if isinstance(region, tuple):  # deal with sets of matched regions
                found = [c for r in region for c in r if c.name in all_centroids]
            else:
                found = [r for r in region if r.name in all_centroids]
            assert len(found) > 0
            result.append(
                tuple(pointset.PointSet(
                    [all_centroids[r.name] for r in found], space=space
                ).centroid)
            )
        return result

    def _arraylike_to_dataframe(self, array: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        """
        Convert a numpy array with the connectivity matrix to
        a DataFrame with regions as column and row headers.
        """
        if not isinstance(array, np.ndarray):
            array = array.to_numpy()
        if not (array == array.T).all():
            logger.warning("The connectivity matrix is not symmetric.")
        df = pd.DataFrame(array)
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parc = next(iter(parcellations))
        with QUIET:
            indexmap = {
                i: parc.get_region(regionname, allow_tuple=True)
                for i, regionname in enumerate(self.regions)
            }
        nrows = array.shape[0]
        try:
            assert len(indexmap) == nrows
            remapper = {
                label - min(indexmap.keys()): region
                for label, region in indexmap.items()
            }
            df = df.rename(index=remapper).rename(columns=remapper)
        except Exception:
            raise RuntimeError("Could not decode connectivity matrix regions.")
        return df

    def _load_matrix(self, subject: str):
        """
        Extract connectivity matrix.
        """
        assert subject in self.subjects

        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parc = next(iter(parcellations))
        with QUIET:
            indexmap = {
                i: parc.get_region(regionname, allow_tuple=True)
                for i, regionname in enumerate(self.regions)
            }
        try:
            #assert len(indexmap) == nrows
            df = self._connector.get(self._files[subject], decode_func=self._decode_func)
            remapper = {
                label - min(indexmap.keys()): region
                for label, region in indexmap.items()
            }
            df.rename(index=remapper, inplace=True)
            df.rename(columns=remapper, inplace=True)
        except Exception:
            raise RuntimeError("Could not decode connectivity matrix regions.")
        return df



    def plot(self, subject: str = None, regions: str = None,
             logscale: bool = False, *args, backend="nilearn", **kwargs
    ):
        """
        Plots the heatmap of the connectivity matrix using nilearn.plotting.

        Parameters
        ----------
        subject: str
            Name of the subject (see ConnectivityMatrix.subjects for available names).
            If "mean" or None is given, the mean is taken in case of multiple
            available matrices.
        regions: list[str]
            Display the matrix only for selected regions. By default, shows all the regions.
            It can only be a subset of regions of the feature.
        logscale: bool
            Display the data in log10 scale
        backend: str
            "nilearn" or "plotly"
        **kwargs:
            Can take all the arguments `nilearn.plotting.plot_matrix` can take. See the doc at
            https://nilearn.github.io/stable/modules/generated/nilearn.plotting.plot_matrix.html
        """
        try:
            if regions is None:
                return None
                #regions = self.regions
            #indices = [self.regions.index(r) for r in regions]
            m = self.get_matrix(subject=subject)
            #matrix = self.get_matrix(subject=subject).iloc[indices, indices].to_numpy()  # nilearn.plotting.plot_matrix works better with a numpy array
            metadata = m.meta['data']
            d = m.rename(str, axis='columns').rename(str, axis='index')

            d = d[d[regions] != 0][[regions]]
            d = d.map(np.log10)
            d.sort_values(by=[regions], inplace=True)
            scale = [
                (0, '#e6e6e6'),
                (0.005, '#f4f78c'),
                (0.13, '#e28b21'),
                (0.28, '#d75223'),
                (0.32, '#d44324'),
                (0.74, '#7c5359'),
                (0.92, '#513331'),
                (1, '#513331')
            ]
        except Exception as e:
            logger.exception('error')
        try:
            d = d.rename(name_to_code, axis='columns').rename(name_to_code, axis='index')
            regions = name_to_code(regions)
        except Exception as e:
            logger.error("error rename column", exc_info=True)
        if backend == "plotly":
            try:
                import plotly.express as px
                #fig = px.bar(d, x=regions, orientation='h', color=regions, color_continuous_scale=scale, range_color=[-6, 0])
                fig = px.scatter(
                    d, x=regions,
                    color=regions, color_continuous_scale=scale, range_color=[-6, 0],
                )
                #fig.update_yaxes(autorange='reversed')
                fig.update_layout(
                    xaxis={'range': [-6, 0], 'title': f'log<sub>10</sub>(FLNe) for injections in area {regions}'},
                    yaxis={
                        'dtick': 1,
                        'tick0': 0,
                        'range': [-1, len(d[regions])],
                        'autorange': False,
                        'title': 'source cortical area'
                    },
                    height=len(d) * 18,
                    coloraxis_showscale=False,
                    font={
                        'family': 'Sans Serif',
                        'size': 10,
                    }
                )
                fig.update_traces(
                    hovertemplate=f'log10(FLNe %{{y}} → {regions}) = %{{x:.2f}}'
                )
                inj = metadata['injections'][regions]
                _i = [f'  • <a href="{i["link"]}" target="_blank">{i["injection"]}</a> (⤓<a href="{i["csv"]}" target="_blank">csv</a>, ⤓<a href="{i.get("cells")}" target="_blank">xyz</a>)  ' for i in inj]
                fig.add_annotation(text=f'Injections into {regions}: <br>{"<br>".join(_i)}', xref='paper', yref='paper', x=0.9, y=0.01, showarrow=False, bordercolor="rgba(0, 0, 0, 1)", bgcolor="rgba(255, 255, 255, 1)", align='left')
            except Exception:
                logger.exception('error')
        return fig



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re
from urllib.parse import urlparse
from typing import Callable, Union
from zipfile import ZipFile
import json
import os

from .. import anchor as _anchor
from . import cortical_profile, tabular

from ...commons import PolyLine, logger, create_key
from ...core import region as _region
from ...retrieval import requests
from ...retrieval.repositories import RepositoryConnector, ZipfileConnector
from ...retrieval.cache import CACHE

from skimage.draw import polygon
from skimage.transform import resize
from io import BytesIO
from textwrap import wrap
import numpy as np
import pandas as pd

class DFWithMeta(pd.DataFrame):
    _metadata = ['meta']
    @property
    def _constructor(self):
        return DFWithMeta


class MarmosetCalbindinDensityProfile(
    tabular.Tabular,
    configuration_folder="features/tabular/corticalprofiles/marmosetcelldensity",
    category='cellular'
):

    DESCRIPTION = (
        "Marmoset Calbindin Cell Density Profile"
    )

    @classmethod
    def decode_meta(cls, spec):
        decoder_spec = spec.get("decoder", {})
        if decoder_spec["@type"].endswith('csv'):
            kwargs = {k: v for k, v in decoder_spec.items() if k != "@type"}
            return lambda b: pd.read_csv(BytesIO(b), **kwargs)
        else:
            return None


    class CorticalProfileConnector(ZipfileConnector):
        class ZipFileLoader:
            """
            Loads a file from the zip archive, but mimics the behaviour
            of cached http requests used in other connectors.
            """
            def __init__(self, zipfile, filename, decode_func, meta=None):
                self.zipfile = zipfile
                self.filename = filename
                self.func = decode_func
                self.cachefile = CACHE.build_filename(zipfile + filename)
                self.meta = meta

            @property
            def cached(self):
                return os.path.isfile(self.cachefile)

            @property
            def data(self):
                container = ZipFile(self.zipfile)
                df = self.func(container.open(self.filename).read())
                if self.meta is not None:
                    df = DFWithMeta(df)
                    df.meta = self.meta
                return df


        def get_loader(self, filename, folder="", decode_func=None):
            """Get a lazy loader for a file, for loading data
            only once loader.data is accessed."""
            meta = self.ZipFileLoader(self.zipfile, 'meta.json', lambda b: json.loads(b)).data
            loader = self.ZipFileLoader(self.zipfile, filename, decode_func, meta)
            return loader


    def __init__(
        self,
        url: str,
        anchor: _anchor.AnatomicalAnchor,
        datasets: list = [],
        prerelease: bool = False,
        id: str = None,
        region: Union[str, _region.Region] = None,
        connector: RepositoryConnector = None,
        decode_func: Callable = None

    ):
        """
        Generate a cell density profile from a URL to a cloud folder
        formatted according to the structure used by Bludau/Dickscheid et al.
        """
        id_template = re.compile(r'marmoset-nencki-monash-template--calbindin-profile-(?P<region_code>.*)')
        cortical_profile.CorticalProfile.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Segmented cell body density",
            unit="detected cells / 0.1mm3",
            anchor=anchor,
            datasets=datasets,
            prerelease=prerelease,
            id=id,
        )
        matches = id_template.match(self._id)
        if matches:
            code = matches.group('region_code')
        else:
            raise ValueError('Malformed feature id %s' % (self.id,))
        self.region = region
        self.region_code = code
        self.file = f'{code}.csv'
        self._step = 0.01
        self._url = url
        self._connector = connector
        self._decode_func = decode_func

    @property
    def key(self):
        assert len(self.species) == 1
        return create_key("{}_{}_{}".format(
            self.id,
            self.species[0]['name'],
            self.region_code
        ))

    @property
    def data(self):
        """
        Returns a matrix as a pandas dataframe.
        -------
        pd.DataFrame
            A square matrix with region names as the column and row names.
        """
        try:
            df = self._connector.get(self.file, decode_func=self._decode_func)
        except Exception:
            raise RuntimeError("Could not decode connectivity matrix regions.")
        return df.copy()

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Plot the profile.

        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            Keyword arguments are passed on to the plot command.
            'layercolor' can be used to specify a color for cortical layer shading.
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.get("title", "\n".join(wrap(self.name, wrapwidth)))

        if backend == "matplotlib":
            raise NotImplemented('matplotlib graph is not supported yet')
        elif backend == "plotly":
            try:
                from plotly.subplots import make_subplots
                import plotly.graph_objects as go
                kwargs["title"] = kwargs["title"].replace("\n", "<br>")
                kwargs["labels"] = {
                    "index": kwargs.pop("xlabel", None) or kwargs.pop("index", "Cortical depth"),
                    "value": kwargs.pop("ylabel", None) or kwargs.pop("value", self._unit)
                }
                #fig = self.data.plot(*args, **kwargs, backend=backend)
                data = self.data
                cases = list(data.case_id.unique())
                fig = make_subplots(rows=1, cols=3, subplot_titles=cases)
                bp = data.meta['value']['layer_breakpoint'][self.region_code]
                for idx, c in enumerate(cases):
                    d = data[data.case_id==c]
                    fig.add_trace(go.Scatter(x=d.density_5, y=d.depth, line_color='#555555', mode='lines', name=c), row=1, col=(idx+1))
                    fig.add_trace(go.Scatter(x=d.density_25, y=d.depth, fill='tonexty', line_color='#555555', fillcolor='#eeeeee', mode='lines', name=c), row=1, col=(idx+1))
                    fig.add_trace(go.Scatter(x=d.density_50, y=d.depth, fill='tonexty', line_color='#000000', fillcolor='#cccccc', mode='lines', name=c), row=1, col=(idx+1))
                    fig.add_trace(go.Scatter(x=d.density_75, y=d.depth, fill='tonexty', line_color='#555555', fillcolor='#cccccc', mode='lines', name=c), row=1, col=(idx+1))
                    fig.add_trace(go.Scatter(x=d.density_95, y=d.depth, fill='tonexty', line_color='#555555', fillcolor='#eeeeee', mode='lines', name=c), row=1, col=(idx+1))

                    bp1 = bp['depth_1_3'][c]
                    bp2 = bp['depth_5_6'][c]
                    fig.add_hrect(
                        y0=0, y1=bp1, line_width=0, fillcolor="gray",
                        opacity=0.,
                        layer='below',
                        row=1, col=(idx+1)
                    )
                    fig.add_hrect(
                        y0=bp1, y1=bp2, line_width=0, fillcolor="gray",
                        opacity=0.2,
                        layer='below',
                        row=1, col=(idx+1)
                    )
                    fig.add_hrect(
                        y0=bp2, y1=1, line_width=0, fillcolor="gray",
                        opacity=0.,
                        layer='below',
                        row=1, col=(idx+1)
                    )
                    if idx == len(cases) - 1:
                        fig.add_hrect(
                            y0=0, y1=bp1, line_width=0, fillcolor="gray",
                            opacity=0.,
                            label=dict(text='<b>I-III</b>', textposition="middle right", font={'family': 'Arial'}),
                            row=1, col=3
                        )
                        fig.add_hrect(
                            y0=bp1, y1=bp2, line_width=0, fillcolor="gray",
                            opacity=0.,
                            label=dict(text='<b>VI</b>', textposition="middle right", font={'family': 'Arial'}),
                            row=1, col=3
                        )
                        fig.add_hrect(
                            y0=bp2, y1=1, line_width=0, fillcolor="gray",
                            opacity=0.,
                            label=dict(text='<b>V-VI</b>', textposition="middle right", font={'family': 'Arial'}),
                            row=1, col=3
                        )
                fig.update_yaxes(autorange='reversed', title='normalized cortical depth', row=1, col=1)
                fig.update_yaxes(autorange='reversed', showticklabels=False, row=1, col=2)
                fig.update_yaxes(autorange='reversed', showticklabels=False, row=1, col=3)
                max_x = math.ceil(data.density_95.max())
                fig.update_xaxes(title='CB+ density<br>(·10<sup>3</sup> ·mm<sup>-3</sup>)', range=[0, max_x], row=1, col=1)
                fig.update_xaxes(title='CB+ density<br>(·10<sup>3</sup> ·mm<sup>-3</sup>)', range=[0, max_x], row=1, col=2)
                fig.update_xaxes(title='CB+ density<br>(·10<sup>3</sup> ·mm<sup>-3</sup>)', range=[0, max_x], row=1, col=3)
                fig.update_annotations(font_family='Arial')

                fig.update_layout(
                    showlegend=False,
                    font=dict(
                        family='Arial'
                    ),
                    xaxis=dict(
                        tickfont=dict(family='Arial')
                    ),
                    yaxis=dict(
                        tickfont=dict(family='Arial')
                    ),
                    #yaxis_range=(0, max(data)),
                    title=dict(
                        #automargin=True,
                        yref="container",
                        xref="container",
                        pad=dict(t=40), xanchor="left", yanchor="top",

                    )
                )
            except Exception:
                logger.info('error', exc_info=True)
            return fig
        else:
            return self.data.plot(*args, **kwargs, backend=backend)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import cortical_profile

from ...locations import point


class BigBrainIntensityProfile(
    cortical_profile.CorticalProfile,
    category='cellular'
):

    DESCRIPTION = (
        "Cortical profiles of BigBrain staining intensities computed by Konrad Wagstyl, "
        "as described in the publication 'Wagstyl, K., et al (2020). BigBrain 3D atlas of "
        "cortical layers: Cortical and laminar thickness gradients diverge in sensory and "
        "motor cortices. PLoS Biology, 18(4), e3000678. "
        "http://dx.doi.org/10.1371/journal.pbio.3000678'."
        "The data is taken from the tutorial at "
        "https://github.com/kwagstyl/cortical_layers_tutorial. Each vertex is "
        "assigned to the regional map when queried."
    )

    def __init__(
        self,
        regionname: str,
        depths: list,
        values: list,
        boundaries: list,
        location: point.Point
    ):
        from ..anchor import AnatomicalAnchor
        anchor = AnatomicalAnchor(
            location=location,
            region=regionname,
            species='Homo sapiens'
        )
        cortical_profile.CorticalProfile.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Modified silver staining",
            anchor=anchor,
            depths=depths,
            values=values,
            unit="staining intensity",
            boundary_positions={
                b: boundaries[b[0]]
                for b in cortical_profile.CorticalProfile.BOUNDARIES
            }
        )
        self.location = location



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Request files with decoders, lazy loading, and caching."""

from .cache import CACHE
from .exceptions import EbrainsAuthenticationError
from ..commons import (
    logger,
    HBP_AUTH_TOKEN,
    KEYCLOAK_CLIENT_ID,
    KEYCLOAK_CLIENT_SECRET,
    siibra_tqdm,
    SIIBRA_USE_LOCAL_SNAPSPOT,
)
from .. import __version__

import json
from zipfile import ZipFile
import requests
import os
from nibabel import Nifti1Image, GiftiImage, streamlines, freesurfer
from skimage import io as skimage_io
import gzip
from io import BytesIO
import urllib.parse
import pandas as pd
import numpy as np
from typing import List, Callable, Any, TYPE_CHECKING
from enum import Enum
from functools import wraps
from time import sleep
import sys
import platform

if platform.system() == "Linux":
    from filelock import FileLock as Lock
else:
    from filelock import SoftFileLock as Lock

if TYPE_CHECKING:
    from .repositories import GitlabConnector

USER_AGENT_HEADER = {"User-Agent": f"siibra-python/{__version__}"}


def read_as_bytesio(function: Callable, suffix: str, bytesio: BytesIO):
    """
    Helper method to provide BytesIO to methods that only takes file path and
    cannot handle BytesIO normally (e.g., `nibabel.freesurfer.read_annot()`).

    Writes the bytes to a temporary file on cache and reads with the
    original function.

    Parameters
    ----------
    function : Callable
    suffix : str
        Must match the suffix expected by the function provided.
    bytesio : BytesIO

    Returns
    -------
    Return type of the provided function.
    """
    tempfile = CACHE.build_filename(f"temp_{suffix}") + suffix
    with open(tempfile, "wb") as bf:
        bf.write(bytesio.getbuffer())
    result = function(tempfile)
    os.remove(tempfile)
    return result


DECODERS = {
    ".nii": lambda b: Nifti1Image.from_bytes(b),
    ".gii": lambda b: GiftiImage.from_bytes(b),
    ".json": lambda b: json.loads(b.decode()),
    ".tck": lambda b: streamlines.load(BytesIO(b)),
    ".csv": lambda b: pd.read_csv(BytesIO(b)),
    ".tsv": lambda b: pd.read_csv(BytesIO(b), delimiter="\t").dropna(axis=0, how="all"),
    ".txt": lambda b: pd.read_csv(BytesIO(b), delimiter=" ", header=None),
    ".zip": lambda b: ZipFile(BytesIO(b)),
    ".png": lambda b: skimage_io.imread(BytesIO(b)),
    ".npy": lambda b: np.load(BytesIO(b)),
    ".annot": lambda b: read_as_bytesio(freesurfer.read_annot, '.annot', BytesIO(b)),
}


def find_suitiable_decoder(url: str) -> Callable:
    """
    By supplying a url or a filename, obtain a suitable decoder function
    for siibra to digest based on predifined DECODERS. An extra layer of 
    gzip decompresser automatically added for gzipped files.

    Parameters
    ----------
    url : str
        The url or filename with extension.

    Returns
    -------
    Callable or None
    """
    urlpath = urllib.parse.urlsplit(url).path
    if urlpath.endswith(".gz"):
        dec = find_suitiable_decoder(urlpath[:-3])
        if dec is None:
            return lambda b: gzip.decompress(b)
        else:
            return lambda b: dec(gzip.decompress(b))

    suitable_decoders = [
        dec for sfx, dec in DECODERS.items() if urlpath.endswith(sfx)
    ]
    if len(suitable_decoders) > 0:
        assert len(suitable_decoders) == 1
        return suitable_decoders[0]
    else:
        return None


class SiibraHttpRequestError(Exception):
    def __init__(self, url: str, status_code: int, msg="Cannot execute http request."):
        self.url = url
        self.status_code = status_code
        self.msg = msg
        Exception.__init__(self)

    def __str__(self):
        return f"{self.msg}\n\tStatus code: {self.status_code}\n\tUrl: {self.url:76.76}"


class HttpRequest:
    def __init__(
        self,
        url: str,
        func: Callable = None,
        msg_if_not_cached: str = None,
        refresh=False,
        post=False,
        **kwargs,
    ):
        """
        Initialize a cached http data loader.
        It takes a URL and optional data conversion function.
        For loading, the http request is only performed if the
        result is not yet available in the disk cache.
        Leaves the interpretation of the returned content to the caller.

        Parameters
        ----------
        url : string, or None
            URL for loading raw data, which is then fed into `func`
            for creating the output.
            If None, `func` will be called without arguments.
        func : function pointer
            Function for constructing the output data
            (called on the data retrieved from `url`, if supplied)
        refresh : bool, default: False
            If True, a possibly cached content will be ignored and refreshed
        post: bool, default: False
            perform a post instead of get
        """
        assert url is not None
        self.url = url
        self._set_decoder_func(func)
        self.kwargs = kwargs
        self.cachefile = CACHE.build_filename(self.url + json.dumps(kwargs))
        self.msg_if_not_cached = msg_if_not_cached
        self.refresh = refresh
        self.post = post

    def _set_decoder_func(self, func: Callable = None):
        """
        Sets the decoder function of the HttpRequest. If `func` is None,
        it will try to find a suitable decoder.

        Parameters
        ----------
        func : Callable, default: None
        """
        self.func = func or find_suitiable_decoder(self.url)

    @property
    def cached(self):
        return os.path.isfile(self.cachefile)

    def _retrieve(self, block_size=1024, min_bytesize_with_no_progress_info=2e8):
        """
        Populates the file cache with the data from http if required.
        noop if 1/ data is already cached and 2/ refresh flag not set
        The caller should load the cachefile after _retrieve successfuly executes
        """
        if self.cached and not self.refresh:
            return

        # not yet in cache, perform http request.
        if self.msg_if_not_cached is not None:
            logger.debug(self.msg_if_not_cached)

        headers = self.kwargs.get("headers", {})
        other_kwargs = {
            key: self.kwargs[key] for key in self.kwargs if key != "headers"
        }

        http_method = requests.post if self.post else requests.get
        r = http_method(
            self.url,
            headers={
                **USER_AGENT_HEADER,
                **headers,
            },
            **other_kwargs,
            stream=True,
        )

        if not r.ok:
            raise SiibraHttpRequestError(status_code=r.status_code, url=self.url)

        size_bytes = int(r.headers.get("content-length", 0))
        if size_bytes > min_bytesize_with_no_progress_info:
            progress_bar = siibra_tqdm(
                total=size_bytes,
                unit="iB",
                unit_scale=True,
                position=0,
                leave=True,
                desc=f"Downloading {os.path.split(self.url)[-1]} ({size_bytes / 1024**2:.1f} MiB)",
            )
        temp_cachefile = f"{self.cachefile}_temp"
        lock = Lock(f"{temp_cachefile}.lock")

        with lock:
            with open(temp_cachefile, "wb") as f:
                for data in r.iter_content(block_size):
                    if size_bytes > min_bytesize_with_no_progress_info:
                        progress_bar.update(len(data))
                    f.write(data)
            if size_bytes > min_bytesize_with_no_progress_info:
                progress_bar.close()
            if self.refresh and os.path.isfile(self.cachefile):
                os.remove(self.cachefile)
            self.refresh = False
            os.rename(temp_cachefile, self.cachefile)

    def get(self):
        self._retrieve()
        with open(self.cachefile, "rb") as f:
            data = f.read()
        try:
            return data if self.func is None else self.func(data)
        except Exception as e:
            # if network error results in bad cache, it may get raised here
            # e.g. BadZipFile("File is not a zip file")
            # if that happens, remove cachefile and
            try:
                os.unlink(self.cachefile)
            except Exception:
                pass
            raise e

    @property
    def data(self):
        # for backward compatibility with old LazyHttpRequest class
        return self.get()


class ZipfileRequest(HttpRequest):
    def __init__(self, url, filename, func=None, refresh=False):
        HttpRequest.__init__(
            self, url, refresh=refresh,
            func=func or find_suitiable_decoder(filename)
        )
        self.filename = filename

    def get(self):
        self._retrieve()
        zipfile = ZipFile(self.cachefile)
        filenames = zipfile.namelist()
        matches = [fn for fn in filenames if fn.endswith(self.filename)]
        if len(matches) == 0:
            raise RuntimeError(
                f"Requested filename {self.filename} not found in archive at {self.url}"
            )
        if len(matches) > 1:
            raise RuntimeError(
                f'Requested filename {self.filename} was not unique in archive at {self.url}. Candidates were: {", ".join(matches)}'
            )
        with zipfile.open(matches[0]) as f:
            data = f.read()
        return data if self.func is None else self.func(data)


class EbrainsRequest(HttpRequest):
    """
    Implements lazy loading of HTTP Knowledge graph queries.
    """

    _KG_API_TOKEN: str = None
    _IAM_ENDPOINT: str = "https://iam.ebrains.eu/auth/realms/hbp"
    _IAM_DEVICE_ENDPOINT: str = None
    _IAM_DEVICE_MAXTRIES = 12
    _IAM_DEVICE_POLLING_INTERVAL_SEC = 5
    _IAM_DEVICE_FLOW_CLIENTID = "siibra"

    keycloak_endpoint = (
        "https://iam.ebrains.eu/auth/realms/hbp/protocol/openid-connect/token"
    )

    def __init__(
        self, url, decoder=None, params={}, msg_if_not_cached=None, post=False
    ):
        """Construct an EBRAINS request."""
        # NOTE: we do not pass params and header here,
        # since we want to evaluate them late in the get() method.
        # This is nice because it allows to set env. variable KG_TOKEN only when
        # really needed, and not necessarily on package initialization.
        self.params = params
        HttpRequest.__init__(self, url, decoder, msg_if_not_cached, post=post)

    @classmethod
    def init_oidc(cls):
        resp = requests.get(f"{cls._IAM_ENDPOINT}/.well-known/openid-configuration")
        json_resp = resp.json()
        if "token_endpoint" in json_resp:
            logger.debug(
                f"token_endpoint exists in .well-known/openid-configuration. Setting _IAM_TOKEN_ENDPOINT to {json_resp.get('token_endpoint')}"
            )
            cls._IAM_TOKEN_ENDPOINT = json_resp.get("token_endpoint")
        else:
            logger.warning(
                "expect token endpoint in .well-known/openid-configuration, but was not present"
            )

        if "device_authorization_endpoint" in json_resp:
            logger.debug(
                f"device_authorization_endpoint exists in .well-known/openid-configuration. setting _IAM_DEVICE_ENDPOINT to {json_resp.get('device_authorization_endpoint')}"
            )
            cls._IAM_DEVICE_ENDPOINT = json_resp.get("device_authorization_endpoint")
        else:
            logger.warning(
                "expected device_authorization_endpoint in .well-known/openid-configuration, but was not present"
            )

    @classmethod
    def fetch_token(cls, **kwargs):
        """
        Fetch an EBRAINS token using commandline-supplied username/password
        using the data proxy endpoint.


        :ref:`Details on how to access EBRAINS are here.<accessEBRAINS>`
        """
        cls.device_flow(**kwargs)

    @classmethod
    def device_flow(cls, **kwargs):
        if all(
            [
                not sys.__stdout__.isatty(),  # if is tty, do not raise
                not any(
                    k in ["JPY_INTERRUPT_EVENT", "JPY_PARENT_PID"] for k in os.environ
                ),  # if is notebook environment, do not raise
                not os.getenv(
                    "SIIBRA_ENABLE_DEVICE_FLOW"
                ),  # if explicitly enabled by env var, do not raise
            ]
        ):
            raise EbrainsAuthenticationError(
                "sys.stdout is not tty, SIIBRA_ENABLE_DEVICE_FLOW is not set,"
                "and not running in a notebook. Are you running in batch mode?"
            )

        cls.init_oidc()

        def get_scope() -> str:
            scope = kwargs.get("scope")
            if not scope:
                return None
            if not isinstance(scope, list):
                logger.warning("scope needs to be a list, is but is not... skipping")
                return None
            if not all(isinstance(scope, str) for scope in scope):
                logger.warning("scope needs to be all str, but is not")
                return None
            if len(scope) == 0:
                logger.warning("provided empty list as scope... skipping")
                return None
            return "+".join(scope)

        scope = get_scope()

        data = {"client_id": cls._IAM_DEVICE_FLOW_CLIENTID}

        if scope:
            data["scope"] = scope

        resp = requests.post(url=cls._IAM_DEVICE_ENDPOINT, data=data)
        resp.raise_for_status()
        resp_json = resp.json()
        logger.debug("device flow, request full json:", resp_json)

        assert "verification_uri_complete" in resp_json
        assert "device_code" in resp_json

        device_code = resp_json.get("device_code")

        print("***")
        print(f"To continue, please go to {resp_json.get('verification_uri_complete')}")
        print("***")

        attempt_number = 0
        sleep_timer = cls._IAM_DEVICE_POLLING_INTERVAL_SEC
        while True:
            # TODO the polling is a little busted at the moment.
            # need to speak to axel to shorten the polling duration
            sleep(sleep_timer)

            logger.debug("Calling endpoint")
            if attempt_number > cls._IAM_DEVICE_MAXTRIES:
                message = (
                    f"exceeded max attempts: {cls._IAM_DEVICE_MAXTRIES}, aborting..."
                )
                logger.error(message)
                raise EbrainsAuthenticationError(message)
            attempt_number += 1
            resp = requests.post(
                url=cls._IAM_TOKEN_ENDPOINT,
                data={
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "client_id": cls._IAM_DEVICE_FLOW_CLIENTID,
                    "device_code": device_code,
                },
            )

            if resp.status_code == 200:
                json_resp = resp.json()
                logger.debug("Device flow sucessful:", json_resp)
                cls._KG_API_TOKEN = json_resp.get("access_token")
                print("ebrains token successfuly set.")
                break

            if resp.status_code == 400:
                json_resp = resp.json()
                error = json_resp.get("error")
                if error == "slow_down":
                    sleep_timer += 1
                logger.debug(f"400 error: {resp.content}")
                continue

            raise EbrainsAuthenticationError(resp.content)

    @classmethod
    def set_token(cls, token):
        logger.info(f"Setting EBRAINS Knowledge Graph authentication token: {token}")
        cls._KG_API_TOKEN = token

    @property
    def kg_token(self):
        # token is available, return it
        if self.__class__._KG_API_TOKEN is not None:
            return self.__class__._KG_API_TOKEN

        # See if a token is directly provided in  $HBP_AUTH_TOKEN
        if HBP_AUTH_TOKEN:
            self.__class__._KG_API_TOKEN = HBP_AUTH_TOKEN
            return self.__class__._KG_API_TOKEN

        # try KEYCLOAK. Requires the following environment variables set:
        # KEYCLOAK_ENDPOINT, KEYCLOAK_CLIENT_ID, KEYCLOAK_CLIENT_SECRET

        if KEYCLOAK_CLIENT_ID is not None and KEYCLOAK_CLIENT_SECRET is not None:
            logger.info("Getting an EBRAINS token via keycloak client configuration...")
            result = requests.post(
                self.__class__._IAM_TOKEN_ENDPOINT,
                data=(
                    f"grant_type=client_credentials&client_id={KEYCLOAK_CLIENT_ID}"
                    f"&client_secret={KEYCLOAK_CLIENT_SECRET}"
                    "&scope=kg-nexus-role-mapping%20kg-nexus-service-account-mock"
                ),
                headers={
                    "content-type": "application/x-www-form-urlencoded",
                    **USER_AGENT_HEADER,
                },
            )
            try:
                content = json.loads(result.content.decode("utf-8"))
            except json.JSONDecodeError as error:
                logger.error(f"Invalid json from keycloak:{error}")
                self.__class__._KG_API_TOKEN = None
            if "error" in content:
                logger.error(content["error_description"])
                self.__class__._KG_API_TOKEN = None
            self.__class__._KG_API_TOKEN = content["access_token"]

        if self.__class__._KG_API_TOKEN is None:
            # No success getting the token
            raise RuntimeError(
                "No access token for EBRAINS Knowledge Graph found. "
                "If you do not have an EBRAINS account, please first register at "
                "https://ebrains.eu/register. Then, use one of the following option: "
                "\n 1. Let siibra get you a token by using siibra.fetch_ebrains_token() and follow the prompt."
                "\n 2. If you know how to get a token yourself, set it as $HBP_AUTH_TOKEN or siibra.set_ebrains_token()"
                "\n 3. If you are an application developer, you might configure keycloak access by setting $KEYCLOAK_CLIENT_ID"
                "and $KEYCLOAK_CLIENT_SECRET."
            )

        return self.__class__._KG_API_TOKEN

    @property
    def auth_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.kg_token}",
        }

    def get(self):
        """Evaluate KG Token is evaluated only on execution of the request."""
        self.kwargs = {"headers": self.auth_headers, "params": self.params}
        return super().get()


def try_all_connectors():
    def outer(fn):
        @wraps(fn)
        def inner(self: "GitlabProxyEnum", *args, **kwargs):
            exceptions = []
            for connector in self.connectors:
                try:
                    return fn(self, *args, connector=connector, **kwargs)
                except Exception as e:
                    exceptions.append(e)
            else:
                for exc in exceptions:
                    logger.error(exc)
                raise Exception("try_all_connectors failed")

        return inner

    return outer


class GitlabProxyEnum(Enum):
    DATASET_V1 = "DATASET_V1"
    PARCELLATIONREGION_V1 = "PARCELLATIONREGION_V1"
    DATASET_V3 = "DATASET_V3"
    DATASETVERSION_V3 = "DATASETVERSION_V3"

    @property
    def connectors(self) -> List["GitlabConnector"]:
        servers = [
            ("https://jugit.fz-juelich.de", 7846),
            ("https://gitlab.ebrains.eu", 421),
        ]
        from .repositories import GitlabConnector, LocalFileRepository

        if SIIBRA_USE_LOCAL_SNAPSPOT:
            logger.info(f"Using localsnapshot at {SIIBRA_USE_LOCAL_SNAPSPOT}")
            return [LocalFileRepository(SIIBRA_USE_LOCAL_SNAPSPOT)]
        return [
            GitlabConnector(server[0], server[1], "master", archive_mode=True)
            for server in servers
        ]

    @try_all_connectors()
    def search_files(
        self,
        folder: str,
        suffix=None,
        recursive=True,
        *,
        connector: "GitlabConnector" = None,
    ) -> List[str]:
        assert connector
        return connector.search_files(folder, suffix=suffix, recursive=recursive)

    @try_all_connectors()
    def get(self, filename, decode_func=None, *, connector: "GitlabConnector" = None):
        assert connector
        return connector.get(filename, "", decode_func)


class GitlabProxy(HttpRequest):
    folder_dict = {
        GitlabProxyEnum.DATASET_V1: "ebrainsquery/v1/dataset",
        GitlabProxyEnum.DATASET_V3: "ebrainsquery/v3/Dataset",
        GitlabProxyEnum.DATASETVERSION_V3: "ebrainsquery/v3/DatasetVersion",
        GitlabProxyEnum.PARCELLATIONREGION_V1: "ebrainsquery/v1/parcellationregions",
    }

    def __init__(
        self,
        flavour: GitlabProxyEnum,
        instance_id=None,
        postprocess: Callable[["GitlabProxy", Any], Any] = (
            lambda proxy, obj: obj
            if hasattr(proxy, "instance_id") and proxy.instance_id
            else {"results": obj}
        ),
    ):
        if flavour not in GitlabProxyEnum:
            raise RuntimeError("Can only proxy enum members")

        self.flavour = flavour
        self.folder = self.folder_dict[flavour]
        self.postprocess = postprocess
        self.instance_id = instance_id
        self._cached_files = None

    def get(self):
        if self.instance_id:
            return self.postprocess(
                self, self.flavour.get(f"{self.folder}/{self.instance_id}.json")
            )
        return self.postprocess(self, self.flavour.get(f"{self.folder}/_all.json"))


class MultiSourceRequestException(Exception):
    pass


class MultiSourcedRequest:
    requests: List[HttpRequest] = []

    def __init__(self, requests: List[HttpRequest]) -> None:
        self.requests = requests

    def get(self):
        exceptions = []
        for req in self.requests:
            try:
                return req.get()
            except Exception as e:
                exceptions.append(e)
        else:
            raise MultiSourceRequestException(
                "All requests failed:\n" + "\n".join(str(exc) for exc in exceptions)
            )

    @property
    def data(self):
        return self.get()



from typing import Optional, TYPE_CHECKING
from urllib.parse import quote_plus
from numpy import int32
import numpy as np
import re
from dataclasses import dataclass
import math

from .util import encode_number, separator, cipher, neg, decode_number, post_process

if TYPE_CHECKING:
    from siibra.core.atlas import Atlas
    from siibra.core.space import Space
    from siibra.locations import BoundingBox, Point
    from siibra.core.parcellation import Parcellation
    from siibra.core.region import Region
    from siibra.features.feature import Feature

class DecodeNavigationException(Exception): pass

min_int32=-2_147_483_648
max_int32=2_147_483_647


default_root_url='https://atlases.ebrains.eu/viewer/'

def sanitize_id(id: str):
    return id.replace('/', ':')

def get_perspective_zoom(atlas: "Atlas", space: "Space", parc: "Parcellation", region: Optional["Region"]):
    import siibra
    if atlas is siibra.atlases['rat'] or atlas is siibra.atlases['mouse']:
        return 200000
    return 2000000

def get_zoom(atlas: "Atlas", space: "Space", parc: "Parcellation", region: Optional["Region"]):
    import siibra
    if atlas is siibra.atlases['rat'] or atlas is siibra.atlases['mouse']:
        return 35000
    return 350000

supported_prefix = (
  "nifti://",
  "swc://",
  "precomputed://",
  "deepzoom://"
)

def append_query_params(url: str, *args, query_params={}, **kwargs):
    query_str = "&".join([f"{key}={quote_plus(value)}" for key, value in query_params.items()])
    if len(query_str) > 0:
        query_str = "?" + query_str
    return url + query_str

@post_process(append_query_params)
def encode_url(atlas: "Atlas", space: "Space", parc: "Parcellation", region: Optional["Region"]=None, *, root_url=default_root_url, external_url:str=None, feature: "Feature"=None, ignore_warning=False, query_params={}):
    
    overlay_url = None
    if external_url:
        assert any([external_url.startswith(prefix) for prefix in supported_prefix]), f"url needs to start with {(' , '.join(supported_prefix))}"
        overlay_url = '/x-overlay-layer:{url}'.format(
            url=external_url.replace("/", "%2F")
        )

    zoom = get_zoom(atlas, space, parc, region)
    pzoom = get_perspective_zoom(atlas, space, parc, region)
    
    zoom_kwargs = {
        "encoded_pzoom": encode_number(pzoom, False),
        "encoded_zoom": encode_number(zoom, False)
    }
    nav_string='/@:0.0.0.-W000.._eCwg.2-FUe3._-s_W.2_evlu..{encoded_pzoom}..{encoded_nav}..{encoded_zoom}'

    return_url='{root_url}#/a:{atlas_id}/t:{template_id}/p:{parc_id}{overlay_url}'.format(
        root_url    = root_url,
        atlas_id    = sanitize_id(atlas.id),
        template_id = sanitize_id(space.id),
        parc_id     = sanitize_id(parc.id),
        overlay_url = overlay_url if overlay_url else "",
    )

    if feature is not None:
        return_url = return_url + f"/f:{sanitize_id(feature.id)}"

    if region is None:
        return return_url + nav_string.format(encoded_nav='0.0.0', **zoom_kwargs)
    
    return_url=f'{return_url}/rn:{get_hash(region.name)}'

    try:
        result_props=region.spatial_props(space, maptype='labelled')
        if len(result_props.components) == 0:
            return return_url + nav_string.format(encoded_nav='0.0.0', **zoom_kwargs)
    except Exception as e:
        print(f'Cannot get_spatial_props {str(e)}')
        if not ignore_warning:
            raise e
        return return_url + nav_string.format(encoded_nav='0.0.0', **zoom_kwargs)

    centroid=result_props.components[0].centroid

    encoded_centroid=separator.join([ encode_number(math.floor(val * 1e6)) for val in centroid ])
    return_url=return_url + nav_string.format(encoded_nav=encoded_centroid, **zoom_kwargs)
    return return_url

@dataclass
class DecodedUrl:
    bounding_box: "BoundingBox"

def decode_url(url: str, vp_length=1000):
    import siibra
    try:
        space_match = re.search(r'/t:(?P<space_id>[^/]+)', url)
        space_id = space_match.group("space_id")
        space_id = space_id.replace(":", "/")
        space = siibra.spaces[space_id]
    except Exception as e:
        raise DecodeNavigationException from e

    nav_match = re.search(r'/@:(?P<navigation_str>.+)/?', url)
    navigation_str = nav_match.group("navigation_str")
    for char in navigation_str:
        assert char in cipher or char in [neg, separator], f"char {char} not in cipher, nor separator/neg"
    
    try:
        ori_enc, pers_ori_enc, pers_zoom_enc, pos_enc, zoomm_enc = navigation_str.split(f"{separator}{separator}")
    except Exception as e:
        raise DecodeNavigationException from e
    
    try:
        x_enc, y_enc, z_enc = pos_enc.split(separator)
        pos = [decode_number(val) for val in [x_enc, y_enc, z_enc]]
        zoom = decode_number(zoomm_enc)

        # zoom = nm/pixel
        pt1 = [(coord - (zoom * vp_length / 2)) / 1e6 for coord in pos]
        pt1 = Point(pt1, space)
        
        pt2 = [(coord + (zoom * vp_length / 2)) / 1e6 for coord in pos]
        pt2 = Point(pt2, space)

    except Exception as e:
        raise DecodeNavigationException from e

    bbx = BoundingBox(pt1, pt2, space)
    return DecodedUrl(bounding_box=bbx)
    
def get_hash(full_string: str):
    return_val=0
    with np.errstate(over="ignore"):
        for char in full_string:
            # overflowing is expected and in fact the whole reason why convert number to int32
            
            # in windows, int32((0 - min_int32) << 5), rather than overflow to wraper around, raises OverflowError
            shifted_5 = int32(
                (return_val - min_int32) if return_val > max_int32 else return_val
            << 5)

            return_val = int32(shifted_5 - return_val + ord(char))
            return_val = return_val & return_val
    hex_val = hex(return_val)
    return hex_val[3:]



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import anchor as _anchor
from . import tabular

from ... import commons, vocabularies
from ...retrieval import requests

import pandas as pd
import numpy as np
from textwrap import wrap
from typing import List


class ReceptorDensityFingerprint(
    tabular.Tabular,
    configuration_folder="features/tabular/fingerprints/receptor",
    category='molecular'
):

    DESCRIPTION = (
        "Fingerprint of densities (in fmol/mg protein) of receptors for classical neurotransmitters "
        "obtained by means of quantitative in vitro autoradiography. The fingerprint provides average "
        "density measurments for different receptors measured in tissue samples from different subjects "
        "together with the corresponding standard deviations. "
    )

    def __init__(
        self,
        tsvfile: str,
        anchor: _anchor.AnatomicalAnchor,
        datasets: list = [],
        prerelease: bool = False,
        id: str = None,
    ):
        """ Generate a receptor fingerprint from a URL to a .tsv file
        formatted according to the structure used by Palomero-Gallagher et al.
        """
        tabular.Tabular.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Neurotransmitter receptor density",
            anchor=anchor,
            data=None,  # lazy loading below
            datasets=datasets,
            prerelease=prerelease,
            id=id,
        )
        self._loader = requests.HttpRequest(tsvfile)

    @property
    def unit(self) -> str:
        return self._loader.data.iloc[:, -1][0]

    @property
    def receptors(self) -> List[str]:
        return list(self.data.index)

    @property
    def neurotransmitters(self) -> List[str]:
        # TODO quite a lot of receptor features have undecipherable symbols, mainly double quoted receptor
        # Likely ill-formed tsv's
        return [
            "{} ({})".format(
                vocabularies.RECEPTOR_SYMBOLS[t]['neurotransmitter']['label'],
                vocabularies.RECEPTOR_SYMBOLS[t]['neurotransmitter']['name'],
            ) if t in vocabularies.RECEPTOR_SYMBOLS else
            f"{t} (undeciphered)"
            for t in self.receptors
        ]

    @property
    def data(self):
        if self._data_cached is None:
            label_col, mean_col, std_col = list(self._loader.data.columns)[:3]
            self._data_cached = pd.DataFrame(
                np.array([
                    self._loader.data[mean_col],
                    self._loader.data[std_col]
                ]).T,
                index=self._loader.data[label_col],
                columns=['mean', 'std']
            )
            self._data_cached.index.name = 'receptor'
        return self._data_cached.copy()

    @classmethod
    def parse_tsv_data(cls, data: dict):
        units = {list(v.values())[3] for v in data.values()}
        labels = list(data.keys())
        assert len(units) == 1
        try:
            mean = [data[_]["density (mean)"] for _ in labels]
            std = [data[_]["density (sd)"] for _ in labels]
        except KeyError as e:
            print(str(e))
            commons.logger.error("Could not parse fingerprint from this dictionary")
        return {
            'unit': next(iter(units)),
            'labels': labels,
            'means': [float(m) if m.isnumeric() else 0 for m in mean],
            'stds': [float(s) if s.isnumeric() else 0 for s in std],
        }

    def polar_plot(self, *args, backend='matplotlib', **kwargs):
        """
        Create a polar plot of the fingerprint.
        backend: str
            "matplotlib" or "plotly"
        """
        if backend == "matplotlib":
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                commons.logger.error("matplotlib not available. Plotting of fingerprints disabled.")
                return None
            from collections import deque

            # default args
            wrapwidth = 40
            y = kwargs.pop("y") if "y" in kwargs else self.data.columns[0]
            yerr = kwargs.pop("yerr") if "yerr" in kwargs else None
            if yerr is None:
                yerr = 'std' if 'std' in self.data.columns else None
            ax = kwargs.pop("ax") if "ax" in kwargs else plt.subplot(111, projection="polar")

            datafield = y or self.data.columns[0]
            if yerr is None and 'std' in self.data.columns:
                yerr = 'std'
            # values = list(self.data[datafield])
            angles = deque(np.linspace(0, 2 * np.pi, self.data.shape[0] + 1)[:-1][::-1])
            angles.rotate(5)
            angles = list(angles)
            # for the values, repeat the first element to have a closed plot
            indices = list(range(self.data.shape[0])) + [0]
            y = list(self.data[datafield].iloc[indices])
            plt.plot(angles + [angles[0]], y, "k-", lw=3, **kwargs)
            if yerr:
                bounds0 = y - self.data[yerr].iloc[indices]
                plt.plot(angles + [angles[0]], bounds0, "k", lw=0.5, **kwargs)
                bounds1 = y + self.data[yerr].iloc[indices]
                plt.plot(angles + [angles[0]], bounds1, "k", lw=0.5, **kwargs)
            ax.set_xticks(angles)
            ax.set_xticklabels([_ for _ in self.data.index])
            ax.set_ylabel(self.unit)
            ax.set_title(
                "\n".join(wrap(f"{self.modality} anchored at {self.anchor._regionspec}", wrapwidth))
            )
            ax.tick_params(pad=9, labelsize=10)
            ax.tick_params(axis="y", labelsize=8)
            plt.tight_layout()
            return ax
        elif backend == "plotly":
            from plotly.express import line_polar
            df = pd.DataFrame(
                {
                    "values": pd.concat(
                        [
                            self.data["mean"],
                            self.data["mean"] - self.data["std"],
                            self.data["mean"] + self.data["std"]
                        ]
                    ),
                    "cat":
                        len(self.data) * ["mean"] +\
                        len(self.data) * ["mean - std"] +\
                        len(self.data) * ["mean + std"]
                }
            )
            return line_polar(
                df, r="values", theta=df.index, color="cat", line_close=True, **kwargs
            )
        else:
            raise NotImplementedError

    def plot(self, *args, **kwargs):
        kwargs['xlabel'] = ""
        return super().plot(*args, **kwargs)



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides spatial representations for parcellations and regions."""

from . import volume as _volume, nifti
from .. import logger, QUIET
from ..commons import (
    MapIndex,
    MapType,
    compare_maps,
    iterate_connected_components,
    clear_name,
    create_key,
    create_gaussian_kernel,
    siibra_tqdm,
    Species,
    CompareMapsResult
)
from ..core import concept, space, parcellation, region as _region
from ..locations import point, pointset
from ..retrieval import requests

import numpy as np
from typing import Union, Dict, List, TYPE_CHECKING, Iterable, Tuple
from scipy.ndimage import distance_transform_edt
from collections import defaultdict
from nibabel import Nifti1Image
from nilearn import image
import pandas as pd
from dataclasses import dataclass, asdict

if TYPE_CHECKING:
    from ..core.region import Region


class ExcessiveArgumentException(ValueError): pass


class InsufficientArgumentException(ValueError): pass


class ConflictingArgumentException(ValueError): pass


class NonUniqueIndexError(RuntimeError): pass


class NoVolumeFound(RuntimeError): pass

@dataclass
class Assignment:
    input_structure: int
    centroid: Union[Tuple[np.ndarray], point.Point] 
    volume: int
    fragment: str
    map_value: np.ndarray


@dataclass
class AssignImageResult(CompareMapsResult, Assignment): pass


class Map(concept.AtlasConcept, configuration_folder="maps"):

    def __init__(
        self,
        identifier: str,
        name: str,
        space_spec: dict,
        parcellation_spec: dict,
        indices: Dict[str, Dict],
        volumes: list = [],
        shortname: str = "",
        description: str = "",
        modality: str = None,
        publications: list = [],
        datasets: list = [],
        prerelease: bool = False,
    ):
        """
        Constructs a new parcellation object.

        Parameters
        ----------
        identifier: str
            Unique identifier of the parcellation
        name: str
            Human-readable name of the parcellation
        space_spec: dict
            Specification of the space (use @id or name fields)
        parcellation_spec: str
            Specification of the parcellation (use @id or name fields)
        indices: dict
            Dictionary of indices for the brain regions.
            Keys are exact region names.
            Per region name, a list of dictionaries with fields "volume" and "label" is expected,
            where "volume" points to the index of the Volume object where this region is mapped,
            and optional "label" is the voxel label for that region.
            For continuous / probability maps, the "label" can be null or omitted.
            For single-volume labelled maps, the "volume" can be null or omitted.
        volumes: list[Volume]
            parcellation volumes
        shortname: str, optional
            Shortform of human-readable name
        description: str, optional
            Textual description of the parcellation
        modality: str, default: None
            Specification of the modality used for creating the parcellation
        publications: list
            List of associated publications, each a dictionary with "doi" and/or "citation" fields
        datasets : list
            datasets associated with this concept
        """
        concept.AtlasConcept.__init__(
            self,
            identifier=identifier,
            name=name,
            species=None,  # inherits species from space
            shortname=shortname,
            description=description,
            publications=publications,
            datasets=datasets,
            modality=modality,
            prerelease=prerelease,
        )

        # Since the volumes might include 4D arrays, where the actual
        # volume index points to a z coordinate, we create subvolume
        # indexers from the given volume provider if 'z' is specified.
        self._indices: Dict[str, List[MapIndex]] = {}
        self.volumes: List[_volume.Volume] = []
        remap_volumes = {}
        # TODO: This assumes knowledge of the preconfigruation specs wrt. z.
        # z to subvolume conversion should probably go to the factory.
        for regionname, indexlist in indices.items():
            k = clear_name(regionname)
            self._indices[k] = []
            for index in indexlist:
                vol = index.get('volume', 0)
                assert vol in range(len(volumes))
                z = index.get('z')
                if (vol, z) not in remap_volumes:
                    if z is None:
                        self.volumes.append(volumes[vol])
                    else:
                        self.volumes.append(_volume.Subvolume(volumes[vol], z))
                    remap_volumes[vol, z] = len(self.volumes) - 1
                self._indices[k].append(
                    MapIndex(volume=remap_volumes[vol, z], label=index.get('label'), fragment=index.get('fragment'))
                )

        # make sure the indices are unique - each map/label pair should appear at most once
        all_indices = sum(self._indices.values(), [])
        seen = set()
        duplicates = {x for x in all_indices if x in seen or seen.add(x)}
        if len(duplicates) > 0:
            logger.warning(f"Non unique indices encountered in {self}: {duplicates}")

        self._space_spec = space_spec
        self._parcellation_spec = parcellation_spec
        self._affine_cached = None
        for v in self.volumes:
            # allow the providers to query their parcellation map if needed
            for p in v._providers.values():
                p.parcellation_map = self
            v._space_spec = space_spec

    @property
    def species(self) -> Species:
        # lazy implementation
        if self._species_cached is None:
            self._species_cached = self.space.species
        return self.space._species_cached

    def get_index(self, region: Union[str, "Region"]):
        """
        Returns the unique index corresponding to the specified region.

        Tip
        ----
        Use find_indices() method for a less strict search returning all matches.

        Parameters
        ----------
        region: str or Region

        Returns
        -------
        MapIndex

        Raises
        ------
        NonUniqueIndexError
            If not unique or not defined in this parcellation map.
        """
        matches = self.find_indices(region)
        if len(matches) > 1:
            # if there is an exact match, we still use it. If not, we cannot proceed.
            regionname = region.name if isinstance(region, _region.Region) else region
            for index, matched_name in matches.items():
                if matched_name == regionname:
                    return index
            raise NonUniqueIndexError(
                f"The specification '{region}' matches multiple mapped "
                f"structures in {str(self)}: {list(matches.values())}"
            )
        elif len(matches) == 0:
            raise NonUniqueIndexError(
                f"The specification '{region}' does not match to any structure mapped in {self}"
            )
        else:
            return next(iter(matches))

    def find_indices(self, region: Union[str, "Region"]):
        """
        Returns the volume/label indices in this map which match the given
        region specification.

        Parameters
        ----------
        region: str or Region

        Returns
        -------
        dict
            - keys: MapIndex
            - values: region name
        """
        if region in self._indices:
            return {
                idx: region
                for idx in self._indices[region]
            }
        regionname = region.name if isinstance(region, _region.Region) else region
        matched_region_names = set(_.name for _ in (self.parcellation.find(regionname)))
        matches = matched_region_names & self._indices.keys()
        if len(matches) == 0:
            logger.warning(f"Region {regionname} not defined in {self}")
        return {
            idx: regionname
            for regionname in matches
            for idx in self._indices[regionname]
        }

    def get_region(self, label: int = None, volume: int = 0, index: MapIndex = None):
        """
        Returns the region mapped by the given index, if any.

        Tip
        ----
        Use get_index() or find_indices() methods to obtain the MapIndex.

        Parameters
        ----------
        label: int, default: None
        volume: int, default: 0
        index: MapIndex, default: None

        Returns
        -------
        Region
            A region object defined in the parcellation map.
        """
        if isinstance(label, MapIndex) and index is None:
            raise TypeError("Specify MapIndex with 'index' keyword.")
        if index is None:
            index = MapIndex(volume, label)
        matches = [
            regionname
            for regionname, indexlist in self._indices.items()
            if index in indexlist
        ]
        if len(matches) == 0:
            logger.warning(f"Index {index} not defined in {self}")
            return None
        elif len(matches) == 1:
            return self.parcellation.get_region(matches[0])
        else:
            # this should not happen, already tested in constructor
            raise RuntimeError(f"Index {index} is not unique in {self}")

    @property
    def space(self):
        for key in ["@id", "name"]:
            if key in self._space_spec:
                return space.Space.get_instance(self._space_spec[key])
        return space.Space(None, "Unspecified space", species=Species.UNSPECIFIED_SPECIES)

    @property
    def parcellation(self):
        for key in ["@id", "name"]:
            if key in self._parcellation_spec:
                return parcellation.Parcellation.get_instance(self._parcellation_spec[key])
        logger.warning(
            f"Cannot determine parcellation of {self.__class__.__name__} "
            f"{self.name} from {self._parcellation_spec}"
        )
        return None

    @property
    def labels(self):
        """
        The set of all label indices defined in this map, including "None" if
        not defined for one or more regions.
        """
        return {d.label for v in self._indices.values() for d in v}

    @property
    def maptype(self) -> MapType:
        if all(isinstance(_, int) for _ in self.labels):
            return MapType.LABELLED
        elif self.labels == {None}:
            return MapType.STATISTICAL
        else:
            raise RuntimeError(
                f"Inconsistent label indices encountered in {self}"
            )

    def __len__(self):
        return len(self.volumes)

    @property
    def regions(self):
        return list(self._indices)

    def fetch(
        self,
        region_or_index: Union[str, "Region", MapIndex] = None,
        *,
        index: MapIndex = None,
        region: Union[str, "Region"] = None,
        **kwargs
    ):
        """
        Fetches one particular volume of this parcellation map.

        If there's only one volume, this is the default, otherwise further
        specification is requested:
        - the volume index,
        - the MapIndex (which results in a regional map being returned)

        You might also consider fetch_iter() to iterate the volumes, or
        compress() to produce a single-volume parcellation map.

        Parameters
        ----------
        region_or_index: str, Region, MapIndex
            Lazy match the specification.
        index: MapIndex
            Explicit specification of the map index, typically resulting
            in a regional map (mask or statistical map) to be returned.
            Note that supplying 'region' will result in retrieving the map index of that region
            automatically.
        region: str, Region
            Specification of a region name, resulting in a regional map
            (mask or statistical map) to be returned.
        **kwargs
            - resolution_mm: resolution in millimeters
            - format: the format of the volume, like "mesh" or "nii"
            - voi: a BoundingBox of interest


            Note
            ----
            Not all keyword arguments are supported for volume formats. Format
            is restricted by available formats (check formats property).

        Returns
        -------
        An image or mesh
        """
        try:
            length = len([arg for arg in [region_or_index, region, index] if arg is not None])
            assert length == 1
        except AssertionError:
            if length > 1:
                raise ExcessiveArgumentException("One and only one of region_or_index, region, index can be defined for fetch")
            # user can provide no arguments, which assumes one and only one volume present

        if isinstance(region_or_index, MapIndex):
            index = region_or_index

        if isinstance(region_or_index, (str, _region.Region)):
            region = region_or_index

        mapindex = None
        if region is not None:
            assert isinstance(region, (str, _region.Region))
            mapindex = self.get_index(region)
        if index is not None:
            assert isinstance(index, MapIndex)
            mapindex = index
        if mapindex is None:
            if len(self) == 1:
                mapindex = MapIndex(volume=0, label=None)
            elif len(self) > 1:
                logger.info(
                    "Map provides multiple volumes and no specification is"
                    "provided. Resampling all volumes to the space."
                )
                resolution = kwargs.get("resolution_mm")
                template = self.space.get_template().fetch(
                    resolution_mm=resolution
                )
                aggregated_volume = np.zeros(template.shape, dtype='uint8')
                for i, region in siibra_tqdm(
                    enumerate(self.regions),
                    unit=" volume", desc="Fetching", total=len(self)
                ):
                    regionlabel = i + 1
                    regionmap = image.resample_to_img(
                        self.fetch(region=region, resolution_mm=resolution),
                        template,
                        interpolation='nearest'
                    )
                    aggregated_volume[regionmap.get_fdata() > 0] = regionlabel
                return Nifti1Image(aggregated_volume, affine=template.affine)
            else:
                raise NoVolumeFound("Map provides no volumes.")

        kwargs_fragment = kwargs.pop("fragment", None)
        if kwargs_fragment is not None:
            if (mapindex.fragment is not None) and (kwargs_fragment != mapindex.fragment):
                raise ConflictingArgumentException(
                    "Conflicting specifications for fetching volume fragment: "
                    f"{mapindex.fragment} / {kwargs_fragment}"
                )
            mapindex.fragment = kwargs_fragment

        if mapindex.volume is None:
            mapindex.volume = 0
        if mapindex.volume >= len(self.volumes):
            raise IndexError(
                f"{self} provides {len(self)} mapped volumes, but #{mapindex.volume} was requested."
            )
        try:
            result = self.volumes[mapindex.volume or 0].fetch(
                fragment=mapindex.fragment, label=mapindex.label, **kwargs
            )
        except requests.SiibraHttpRequestError as e:
            print(str(e))

        if result is None:
            raise RuntimeError(f"Error fetching {mapindex} from {self} as {kwargs.get('format', f'{self.formats}')}.")
        return result

    def fetch_iter(self, **kwargs):
        """
        Returns an iterator to fetch all mapped volumes sequentially.

        All arguments are passed on to function Map.fetch().
        """
        fragment = kwargs.pop('fragment') if 'fragment' in kwargs else None
        return (
            self.fetch(
                index=MapIndex(volume=i, label=None, fragment=fragment), **kwargs
            )
            for i in range(len(self))
        )

    @property
    def provides_image(self):
        return any(v.provides_image for v in self.volumes)

    @property
    def fragments(self):
        return {
            index.fragment
            for indices in self._indices.values()
            for index in indices
            if index.fragment is not None
        }

    @property
    def provides_mesh(self):
        return any(v.provides_mesh for v in self.volumes)

    @property
    def formats(self):
        return {f for v in self.volumes for f in v.formats}

    @property
    def is_labelled(self):
        return self.maptype == MapType.LABELLED

    @property
    def affine(self):
        if self._affine_cached is None:
            # we compute the affine from a volumetric volume provider
            for fmt in _volume.Volume.SUPPORTED_FORMATS:
                if fmt not in _volume.Volume.MESH_FORMATS:
                    if fmt not in self.formats:
                        continue
                    try:
                        self._affine_cached = self.fetch(index=MapIndex(volume=0), format=fmt).affine
                        break
                    except Exception:
                        logger.debug("Caught exceptions:\n", exc_info=1)
                        continue
            else:
                raise RuntimeError(f"No volumetric provider in {self} to derive the affine matrix.")
        if not isinstance(self._affine_cached, np.ndarray):
            logger.error("invalid affine:", self._affine_cached)
        return self._affine_cached

    def __iter__(self):
        return self.fetch_iter()

    def compress(self, **kwargs):
        """
        Converts this map into a labelled 3D parcellation map, obtained by
        taking the voxelwise maximum across the mapped volumes and fragments,
        and re-labelling regions sequentially.

        Paramaters
        ----------
        **kwargs: Takes the fetch arguments of its space's template.

        Returns
        -------
        parcellationmap.Map
        """
        if len(self.volumes) == 1 and not self.fragments:
            raise RuntimeError("The map cannot be merged since there are no multiple volumes or fragments.")

        # initialize empty volume according to the template
        template = self.space.get_template().fetch(**kwargs)
        result_data = np.zeros_like(np.asanyarray(template.dataobj))
        voxelwise_max = np.zeros_like(result_data)
        result_nii = Nifti1Image(result_data, template.affine)
        interpolation = 'nearest' if self.is_labelled else 'linear'
        next_labelindex = 1
        region_indices = defaultdict(list)

        for volidx in siibra_tqdm(
            range(len(self.volumes)), total=len(self.volumes), unit='maps',
            desc=f"Compressing {len(self.volumes)} {self.maptype.name.lower()} volumes into single-volume parcellation",
            disable=(len(self.volumes) == 1)
        ):
            for frag in siibra_tqdm(
                self.fragments, total=len(self.fragments), unit='maps',
                desc=f"Compressing {len(self.fragments)} {self.maptype.name.lower()} fragments into single-fragment parcellation",
                disable=(len(self.fragments) == 1 or self.fragments is None)
            ):
                mapindex = MapIndex(volume=volidx, fragment=frag)
                img = self.fetch(mapindex)
                if np.linalg.norm(result_nii.affine - img.affine) > 1e-14:
                    logger.debug(f"Compression requires to resample volume {volidx} ({interpolation})")
                    img = image.resample_to_img(img, result_nii, interpolation)
                img_data = np.asanyarray(img.dataobj)

                if self.is_labelled:
                    labels = set(np.unique(img_data)) - {0}
                else:
                    labels = {None}

                for label in labels:
                    with QUIET:
                        mapindex.__setattr__("label", int(label))
                        region = self.get_region(index=mapindex)
                    if region is None:
                        logger.warning(f"Label index {label} is observed in map volume {self}, but no region is defined for it.")
                        continue
                    region_indices[region.name].append({"volume": 0, "label": next_labelindex})
                    if label is None:
                        update_voxels = (img_data > voxelwise_max)
                    else:
                        update_voxels = (img_data == label)
                    result_data[update_voxels] = next_labelindex
                    voxelwise_max[update_voxels] = img_data[update_voxels]
                    next_labelindex += 1

        return Map(
            identifier=f"{create_key(self.name)}_compressed",
            name=f"{self.name} compressed",
            space_spec=self._space_spec,
            parcellation_spec=self._parcellation_spec,
            indices=region_indices,
            volumes=[
                _volume.Volume(
                    space_spec=self._space_spec,
                    providers=[nifti.NiftiProvider(result_nii)]
                )
            ]
        )

    def compute_centroids(self) -> Dict[str, point.Point]:
        """
        Compute a dictionary of the centroids of all regions in this map.

        Returns
        -------
        Dict[str, point.Point]
            Region names as keys and computed centroids as items.
        """
        centroids = {}
        maparr = None
        for regionname, indexlist in siibra_tqdm(
            self._indices.items(), unit="regions", desc="Computing centroids"
        ):
            assert len(indexlist) == 1
            index = indexlist[0]
            if index.label == 0:
                continue
            with QUIET:
                mapimg = self.fetch(index=index)  # returns a mask of the region
            maparr = np.asanyarray(mapimg.dataobj)
            centroid_vox = np.mean(np.where(maparr == 1), axis=1)
            assert regionname not in centroids
            centroids[regionname] = point.Point(
                np.dot(mapimg.affine, np.r_[centroid_vox, 1])[:3], space=self.space
            )
        return centroids

    def get_resampled_template(self, **fetch_kwargs) -> Nifti1Image:
        """
        Resample the reference space template to fetched map image. Uses
        nilearn.image.resample_to_img to resample the template.

        Parameters
        ----------
        **fetch_kwargs: takes the arguments of Map.fetch()

        Returns
        -------
        Nifti1Image
        """
        source_template = self.space.get_template().fetch()
        map_image = self.fetch(**fetch_kwargs)
        return image.resample_to_img(
            source_template,
            map_image,
            interpolation='continuous'
        )

    def colorize(self, values: dict, **kwargs) -> Nifti1Image:
        """Colorize the map with the provided regional values.

        Parameters
        ----------
        values : dict
            Dictionary mapping regions to values

        Return
        ------
        Nifti1Image
        """

        result = None
        for volidx, vol in enumerate(self.fetch_iter(**kwargs)):
            if isinstance(vol, dict):
                raise NotImplementedError("Map colorization not yet implemented for meshes.")
            img = np.asanyarray(vol.dataobj)
            maxarr = np.zeros_like(img)
            for r, value in values.items():
                index = self.get_index(r)
                if index.volume != volidx:
                    continue
                if result is None:
                    result = np.zeros_like(img)
                    affine = vol.affine
                if index.label is None:
                    updates = img > maxarr
                    result[updates] = value
                    maxarr[updates] = img[updates]
                else:
                    result[img == index.label] = value

        return Nifti1Image(result, affine)

    def get_colormap(self, region_specs: Iterable = None):
        """
        Generate a matplotlib colormap from known rgb values of label indices.

        Parameters
        ----------
        region_specs: iterable(regions), optional
            Optional parameter to only color the desired regions.

        Returns
        -------
        ListedColormap
        """
        from matplotlib.colors import ListedColormap
        import numpy as np

        colors = {}
        if region_specs is not None:
            include_region_names = {
                self.parcellation.get_region(region_spec).name for region_spec in region_specs
            }
        else:
            include_region_names = None

        for regionname, indices in self._indices.items():
            for index in indices:
                if index.label is None:
                    continue

                if (include_region_names is not None) and (regionname not in include_region_names):
                    continue
                else:
                    region = self.get_region(index=index)
                    if region.rgb is not None:
                        colors[index.label] = region.rgb

        pallette = np.array(
            [
                list(colors[i]) + [1] if i in colors else [0, 0, 0, 0]
                for i in range(max(colors.keys()) + 1)
            ]
        ) / [255, 255, 255, 1]
        return ListedColormap(pallette)

    def sample_locations(self, regionspec, numpoints: int):
        """ Sample 3D locations inside a given region.

        The probability distribution is approximated from the region mask based
        on the squared distance transform.

        Parameters
        ----------
        regionspec: Region or str
            Region to be used
        numpoints: int
            Number of samples to draw

        Returns
        -------
        PointSet
            Sample points in physcial coordinates corresponding to this
            parcellationmap
        """
        index = self.get_index(regionspec)
        mask = self.fetch(index=index)
        arr = np.asanyarray(mask.dataobj)
        if arr.dtype.char in np.typecodes['AllInteger']:
            # a binary mask - use distance transform to get sampling weights
            W = distance_transform_edt(np.asanyarray(mask.dataobj))**2
        else:
            # a statistical map - interpret directly as weights
            W = arr
        p = (W / W.sum()).ravel()
        XYZ_ = np.array(
            np.unravel_index(np.random.choice(len(p), numpoints, p=p), W.shape)
        ).T
        XYZ = np.dot(mask.affine, np.c_[XYZ_, np.ones(numpoints)].T)[:3, :].T
        return pointset.PointSet(XYZ, space=self.space)

    def to_sparse(self):
        """
        Creates a SparseMap object from this parcellation map object.

        Returns
        -------
        SparseMap
        """
        from .sparsemap import SparseMap
        indices = {
            regionname: [
                {'volume': idx.volume, 'label': idx.label, 'fragment': idx.fragment}
                for idx in indexlist
            ]
            for regionname, indexlist in self._indices.items()
        }
        return SparseMap(
            identifier=self.id,
            name=self.name,
            space_spec={'@id': self.space.id},
            parcellation_spec={'@id': self.parcellation.id},
            indices=indices,
            volumes=self.volumes,
            shortname=self.shortname,
            description=self.description,
            modality=self.modality,
            publications=self.publications,
            datasets=self.datasets
        )

    def _read_voxel(
        self,
        x: Union[int, np.ndarray, List],
        y: Union[int, np.ndarray, List],
        z: Union[int, np.ndarray, List]
    ):
        def _read_voxels_from_volume(xyz, volimg):
            xyz = np.stack(xyz, axis=1)
            valid_points_mask = np.all([(0 <= di) & (di < vol_size) for vol_size, di in zip(volimg.shape, xyz.T)], axis=0)
            x, y, z = xyz[valid_points_mask].T
            valid_points_indices, *_ = np.where(valid_points_mask)
            valid_data_points = np.asanyarray(volimg.dataobj)[x, y, z]
            return zip(valid_points_indices, valid_data_points)

        # integers are just single-element arrays, cast to avoid an extra code branch for integers
        x, y, z = [np.array(di) for di in (x, y, z)]

        fragments = self.fragments or {None}
        return [
            (pointindex, volume, fragment, data_point)
            for fragment in fragments
            for volume, volimg in enumerate(self.fetch_iter(fragment=fragment))
            # transformations or user input might produce points outside the volume, filter these out.
            for (pointindex, data_point) in _read_voxels_from_volume((x, y, z), volimg)
        ]

    def _assign(
        self,
        item: Union[point.Point, pointset.PointSet, Nifti1Image],
        minsize_voxel=1,
        lower_threshold=0.0,
        **kwargs
    ) -> List[Union[Assignment,AssignImageResult]]:
        """
        For internal use only. Returns a dataclass, which provides better static type checking.
        """

        if isinstance(item, point.Point):
            return self._assign_points(pointset.PointSet([item], item.space, sigma_mm=item.sigma), lower_threshold)
        if isinstance(item, pointset.PointSet):
            return self._assign_points(item, lower_threshold)
        if isinstance(item, Nifti1Image):
            return self._assign_image(item, minsize_voxel, lower_threshold, **kwargs)
        
        raise RuntimeError(
            f"Items of type {item.__class__.__name__} cannot be used for region assignment."
        )

    def assign(
        self,
        item: Union[point.Point, pointset.PointSet, Nifti1Image],
        minsize_voxel=1,
        lower_threshold=0.0,
        **kwargs
    ):
        """Assign an input image to brain regions.

        The input image is assumed to be defined in the same coordinate space
        as this parcellation map.

        Parameters
        ----------
        item: Point, PointSet, Nifti1Image
            A spatial object defined in the same physical reference space as
            this parcellation map, which could be a point, set of points, or
            image. If it is an image, it will be resampled to the same voxel
            space if its affine transformation differs from that of the
            parcellation map. Resampling will use linear interpolation for float
            image types, otherwise nearest neighbor.
        minsize_voxel: int, default: 1
            Minimum voxel size of image components to be taken into account.
        lower_threshold: float, default: 0
            Lower threshold on values in the statistical map. Values smaller
            than this threshold will be excluded from the assignment computation.

        Returns
        -------
        assignments: pandas.DataFrame
            A table of associated regions and their scores per component found
            in the input image, or per coordinate provided. The scores are:

                - Value: Maximum value of the voxels in the map covered by an
                input coordinate or input image signal component.
                - Pearson correlation coefficient between the brain region map
                and an input image signal component (NaN for exact coordinates)
                - Contains: Percentage of the brain region map contained in an
                input image signal component, measured from their binarized
                masks as the ratio between the volume of their intersection
                and the volume of the brain region (NaN for exact coordinates)
                - Contained: Percentage of an input image signal component
                contained in the brain region map, measured from their binary
                masks as the ratio between the volume of their intersection and
                the volume of the input image signal component (NaN for exact
                coordinates)
        components: Nifti1Image or None
            If the input was an image, this is a labelled volume mapping the
            detected components in the input image, where pixel values correspond
            to the "component" column of the assignment table. If the input was
            a Point or PointSet, returns None.
        """

        assignments = self._assign(item, minsize_voxel, lower_threshold, **kwargs)

        # format assignments as pandas dataframe
        columns = [
            "input structure",
            "centroid",
            "volume",
            "fragment",
            "region",
            "correlation",
            "intersection over union",
            "map value",
            "map weighted mean",
            "map containedness",
            "input weighted mean",
            "input containedness"
        ]
        if len(assignments) == 0:
            return pd.DataFrame(columns=columns)
        # determine the unique set of observed indices in order to do region lookups
        # only once for each map index occuring in the point list
        labelled = self.is_labelled  # avoid calling this in a loop
        observed_indices = {  # unique set of observed map indices. NOTE: len(observed_indices) << len(assignments)
            (
                a.volume,
                a.fragment,
                a.map_value if labelled else None
            )
            for a in assignments
        }
        region_lut = {  # lookup table of observed region objects
            (v, f, l): self.get_region(
                index=MapIndex(
                    volume=int(v),
                    label=l if l is None else int(l),
                    fragment=f
                )
            )
            for v, f, l in observed_indices
        }

        dataframe_list = []
        for a in assignments:
            item_to_append = {
                "input structure": a.input_structure,
                "centroid": a.centroid,
                "volume": a.volume,
                "fragment": a.fragment,
                "region": region_lut[
                    a.volume,
                    a.fragment,
                    a.map_value if labelled else None
                ],
            }
            # because AssignImageResult is a subclass of Assignment
            # need to check for isinstance AssignImageResult first
            if isinstance(a, AssignImageResult):
                item_to_append = {
                    **item_to_append,
                    **{
                        "correlation": a.correlation,
                        "intersection over union": a.intersection_over_union,
                        "map value": a.map_value,
                        "map weighted mean": a.weighted_mean_of_first,
                        "map containedness": a.intersection_over_first,
                        "input weighted mean": a.weighted_mean_of_second,
                        "input containedness": a.intersection_over_second,
                    }
                }
            elif isinstance(a, Assignment):
                item_to_append = {
                    **item_to_append,
                    **{
                        "correlation": None,
                        "intersection over union": None,
                        "map value": a.map_value,
                        "map weighted mean": None,
                        "map containedness": None,
                        "input weighted mean": None,
                        "input containedness": None,
                    }
                }
            else:
                raise RuntimeError("assignments must be of type Assignment or AssignImageResult!")

            dataframe_list.append(item_to_append)
        df = pd.DataFrame(dataframe_list)
        return (
            df
            .convert_dtypes()  # convert will guess numeric column types
            .reindex(columns=columns)
        )

    def _assign_points(self, points: pointset.PointSet, lower_threshold: float) -> List[Assignment]:
        """
        assign a PointSet to this parcellation map.

        Parameters:
        -----------
        lower_threshold: float, default: 0
            Lower threshold on values in the statistical map. Values smaller than
            this threshold will be excluded from the assignment computation.
        """
        assignments = []

        if points.space != self.space:
            logger.info(
                f"Coordinates will be converted from {points.space.name} "
                f"to {self.space.name} space for assignment."
            )
        # convert sigma to voxel coordinates
        scaling = np.array(
            [np.linalg.norm(self.affine[:, i]) for i in range(3)]
        ).mean()
        phys2vox = np.linalg.inv(self.affine)

        # if all points have the same sigma, and lead to a standard deviation
        # below 3 voxels, we are much faster with a multi-coordinate readout.
        if points.has_constant_sigma:
            sigma_vox = points.sigma[0] / scaling
            if sigma_vox < 3:
                pts_warped = points.warp(self.space.id)
                X, Y, Z = (np.dot(phys2vox, pts_warped.homogeneous.T) + 0.5).astype("int")[:3]
                for pointindex, vol, frag, value in self._read_voxel(X, Y, Z):
                    if value > lower_threshold:
                        position = pts_warped[pointindex].coordinate
                        assignments.append(
                            Assignment(
                                input_structure=pointindex,
                                centroid=tuple(np.array(position).round(2)),
                                volume=vol,
                                fragment=frag,
                                map_value=value
                            )
                        )
                return assignments

        # if we get here, we need to handle each point independently.
        # This is much slower but more precise in dealing with the uncertainties
        # of the coordinates.
        for pointindex, pt in siibra_tqdm(
            enumerate(points.warp(self.space.id)),
            total=len(points), desc="Warping points",
        ):
            sigma_vox = pt.sigma / scaling
            if sigma_vox < 3:
                # voxel-precise - just read out the value in the maps
                N = len(self)
                logger.debug(f"Assigning coordinate {tuple(pt)} to {N} maps")
                x, y, z = (np.dot(phys2vox, pt.homogeneous) + 0.5).astype("int")[:3]
                values = self._read_voxel(x, y, z)
                for _, vol, frag, value in values:
                    if value > lower_threshold:
                        assignments.append(
                            Assignment(
                                input_structure=pointindex,
                                centroid=tuple(pt),
                                volume=vol,
                                fragment=frag,
                                map_value=value
                            )
                        )
            else:
                logger.info(
                    f"Assigning uncertain coordinate {tuple(pt)} to {len(self)} maps."
                )
                kernel = create_gaussian_kernel(sigma_vox, 3)
                r = int(kernel.shape[0] / 2)  # effective radius
                xyz_vox = (np.dot(phys2vox, pt.homogeneous) + 0.5).astype("int")
                shift = np.identity(4)
                shift[:3, -1] = xyz_vox[:3] - r
                # build niftiimage with the Gaussian blob,
                # then recurse into this method with the image input
                W = Nifti1Image(dataobj=kernel, affine=np.dot(self.affine, shift))
                for entry in self._assign(W, lower_threshold=lower_threshold):
                    entry.input_structure = pointindex
                    entry.centroid = tuple(pt)
                    assignments.append(entry)
        return assignments

    def _assign_image(self, queryimg: Nifti1Image, minsize_voxel: int, lower_threshold: float, split_components: bool = True) -> List[AssignImageResult]:
        """
        Assign an image volume to this parcellation map.

        Parameters:
        -----------
        minsize_voxel: int, default: 1
            Minimum voxel size of image components to be taken into account.
        lower_threshold: float, default: 0
            Lower threshold on values in the statistical map. Values smaller than
            this threshold will be excluded from the assignment computation.
        """
        assignments = []

        def resample(img: Nifti1Image, affine: np.ndarray, shape: tuple):
            # resample query image into this image's voxel space, if required
            if (img.affine - affine).sum() == 0:
                return img
            else:
                interp = "nearest" \
                    if issubclass(np.asanyarray(img.dataobj).dtype.type, np.integer) \
                    else "linear"
                return image.resample_img(
                    img,
                    target_affine=affine,
                    target_shape=shape,
                    interpolation=interp,
                )

        def progress(it, N: int = None, desc: str = "", min_elements=5):
            # wraps a progress indicator around the given iterator,
            # but only if the sequence is long.
            seqlen = N or len(it)
            return iter(it) if seqlen < min_elements \
                else siibra_tqdm(it, desc=desc, total=N)
        
        iter_func = iterate_connected_components if split_components \
            else lambda img: [(1, img)]

        with QUIET and _volume.SubvolumeProvider.UseCaching():
            for frag in self.fragments or {None}:
                for vol, vol_img in progress(
                    enumerate(self.fetch_iter(fragment=frag)),
                    N=len(self),
                    desc=f"Assigning to {len(self)} volumes"
                ):
                    queryimg_res = resample(queryimg, vol_img.affine, vol_img.shape)
                    for mode, maskimg in iter_func(queryimg_res):
                        vol_data = np.asanyarray(vol_img.dataobj)
                        position = np.array(np.where(maskimg.get_fdata())).T.mean(0)
                        labels = {v.label for L in self._indices.values() for v in L if v.volume == vol}
                        for label in progress(
                            labels,
                            desc=f"Assigning to {len(labels)} labelled structures"
                        ):
                            targetimg = vol_img if label is None \
                                else Nifti1Image((vol_data == label).astype('uint8'), vol_img.affine)
                            scores = compare_maps(maskimg, targetimg)
                            if scores.intersection_over_union > 0:
                                assignments.append(
                                    AssignImageResult(
                                        input_structure=mode,
                                        centroid=tuple(position.round(2)),
                                        volume=vol,
                                        fragment=frag,
                                        map_value=label,
                                        **asdict(scores)
                                    )
                                )

        return assignments
